
#ifndef lint
static char sccsid[] = "@(#)dirlist.c 3.2 3/20/90 Arthur van Hoff (Turing Institute)";
#endif lint

/**********************************************************************
 * file:        dirlist.c             (c)1990 Turing Institute
 * author:      Arthur van Hoff
 *
 * Description:
 * 	A simple program which can list a directory for a stack.
 *	Used by the OpenFile and SaveFile stacks.
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>

#include "hypernews.h"

hn_any *stack;

#define MAXFILES	1024
#define MAXPATHLEN	512
#define MAXPAT		128
#define MAXPATS		16

int dirlistlen = 0;
char *dirlist[MAXFILES];
static struct stat sb;

#define is_dir()	((sb.st_mode & S_IFDIR) == S_IFDIR)

int npatterns = 0;
char *patterns[MAXPATS];
char pattern[MAXPAT];

void set_pattern(pat)
char *pat;
{
	npatterns = 0;
	strcpy(pattern,pat);
	pat = pattern;
	while (*pat) {
		patterns[npatterns++] = pat;
		for (; *pat && (*pat != '|') ; pat++);
		if (*pat == '|')
			*pat++ = '\0';
	}
}

int match_string(str)
char *str;
{
	int n, len;
	if (!npatterns)
		return 1;
	for (n = 0 ; n < npatterns ; n++) {
		len = strlen(patterns[n]);
		if (!strcmp(patterns[n],&str[strlen(str) - len]))
			return 1;
	}
	return 0;
}

char *filetype()
{
	if ((sb.st_mode & S_IFDIR) == S_IFDIR)
		return "/";
	if ((sb.st_mode & S_IEXEC) == S_IEXEC)
		return "*";
	return "";
}

int filestat(dirname,filename)
char *dirname, *filename;
{
	static char fnm[MAXPATHLEN];
	sprintf(fnm,"%s/%s",dirname,filename);
	if (stat(fnm,&sb) == -1)
		return 0;
	return 1;
}

int ls(dirname)
char *dirname;
{
	int		len = strlen(dirname);
	DIR		*dirp;
	struct dirent 	*dp;

	while (dirlistlen)
		free(dirlist[--dirlistlen]);

	if (!(dirp = opendir(dirname)))
		return 0;
	for (dp = readdir(dirp) ; (dp != NULL) ; dp = readdir(dirp)) {
		if ((!strcmp(dp->d_name,".")) || (!strcmp(dp->d_name,"..")))
			continue;
		if (!filestat(dirname,dp->d_name))
			continue;
		if (is_dir() || match_string(dp->d_name)) {
			char fnm[MAXPATHLEN];
			sprintf(fnm,"%s%s",dp->d_name,filetype());
			dirlist[dirlistlen] = (char*)malloc(strlen(fnm)+1);
			strcpy(dirlist[dirlistlen++],fnm);
		}
	}
	closedir(dirp);
	return 1;
}

void pathname(path,str)
char *path, *str;
{
	int len;

	while (!strncmp("//",str,2))
		str++;
	if (!strncmp("~/",str,2))
		sprintf(path,"%s/%s",getenv("HOME"),str);
	else
		strcpy(path,str);
	while (1) {
		len = strlen(path);
		if ((len >= 1) && (!strncmp(&path[len-1],"/",1)))
			path[len -= 1] = '\0';
		else if ((len >= 2) && (!strncmp(&path[len-2],"/.",2)))
			path[len -= 2] = '\0';
		else if ((len >= 3) && (!strncmp(&path[len-3],"/..",3))) {
			len -= 3;
			while ((len > 1) && (path[len-1] != '/')) len--;
			path[len] = '\0';
		} else {
			if (len == 0)
				strcpy(path,"/");
			return;
		}
	}
	
}

int hnsetpat(any,argc,argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	int n;
	hn_any *files, *mess;
	char thedir[MAXPATHLEN];

	if (argc != 1)
		return 0;
	if (argv[0]->type != hn_string_type)
		return 0;
	set_pattern(argv[0]->u.string);
	return 1;
}

int hnls(any,argc,argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	int n;
	hn_any *files, *mess;
	char thedir[MAXPATHLEN];

	if (argc != 1)
		return 0;
	if (argv[0]->type != hn_string_type)
		return 0;
	pathname(thedir,argv[0]->u.string);
	if (!ls(thedir)) {
		hn_send(any->u.message.target,"DirError",
			hn_new_array(hn_new_string("io error"),NULL));
		return 1;
	}

	files =  hn_new_array_body(dirlistlen);
	for (n = 0 ; n < dirlistlen ; n++)
		files->u.array.argv[n] = hn_new_string(dirlist[n]);

	mess = hn_new_array(hn_new_string(thedir),files,NULL);
	hn_send(any->u.message.target,"SetDirList",mess);
	hn_free(mess);
	return 1;
}

main(argc,argv)
int argc;
char *argv[];
{
	if (argc != 2) {
		fprintf(stderr,"usage: %s stackname",argv[0]);
		exit(1);
	}
	set_pattern("");
	hn_start(argv[0]);

	stack = hn_new_object(NULL,hn_Stack,argv[1]);
	hn_register(hnls,
		hn_new_message(ANY,"ListDir",hn_new_array_body(1)));
	hn_register(hnsetpat,
		hn_new_message(ANY,"SetPattern",hn_new_array_body(1)));
	hn_connect(stack);
	hn_listen(hn_forever);
	hn_stop();
}
