
/* Copyright (c) The Turing Institute 1989 */

#ifndef lint
static char sccsid[] = "@(#)hnutils.c 3.1 3/19/90 Tim Niblett (Turing Institute)";
#endif lint

#include <stdio.h>
#include <strings.h>
#include <hypernews.h>
#include <malloc.h>

#include "hnutils.h"
#include "hash.h"

/* HyperNeWS utilities for the filer */


/*
 * Create entries for EditText from array of strings (null) terminated
 */
hn_any *hn_new_textmess(nlines)
int nlines;
{
	hn_any *arg, *arglist;

	if (!(arg = hn_new_array_body(1)))
		return(hn_null);
	arglist = hn_new_array_body(nlines);
	arg->u.array.argv[0] = arglist;
	return(arg);
}

/*
 * Get single value string from Message into string 
 */
char *hn_get_text_line(text)
hn_any *text;
{
	hn_any *arr;

	if (!text) {
		hn_error("hn_get_text: Null hn_any\n");
		return(0);
	}
	if (text->type != hn_object_type ||
	    text->u.object.type != hn_EditText) {
		hn_error("hn_get_text: EditText expected\n");
		return((char *)NULL);
	}
	arr = hn_get(text, "Value");
	if (arr->type != hn_array_type) {
		hn_error("hn_get_text: array expected\n");
		return((char *)NULL);
	}
	if (arr->u.array.argc == 0) {
		hn_error("hnfi_get_text: array empty\n");
		return((char *)NULL);	
	}
	if (arr->u.array.argv[0]->type != hn_string_type) {
		hn_error("hn_get_text: bad arg type\n");
		return((char *)NULL);
	}
	else {
		return(arr->u.array.argv[0]->u.string);
	}
}

int hn_set_text_line(text, string)
hn_any *text;
char *string;
{
	hn_any *arr;

	if (!text) {
		hn_error("hn_get_text: Null hn_any\n");
		return(0);
	}
	if (text->type != hn_object_type ||
	    text->u.object.type != hn_EditText) {
		return(0);
	}
	if (!(arr = hn_new_textmess(1)))
		return(0);
	arr->u.array.argv[0]->u.array.argv[0] = hn_new_string(string);
	return(hn_send(text, "SetValue", arr));
}


/*
 *
 */
void object_registration(stack, registry)
char *stack;
struct hn_objreg *registry;
{
	hn_any *any;

	while (registry->name) {
		if (registry->type == hn_Stack)
			any = hn_new_object(
				(char *)NULL,
				registry->type,
				registry->name
				);
		else
			any = hn_new_object(
				stack,
				registry->type,
				registry->name
				);
		(void)HN_HASH(registry->name, any);
		registry++;
	}
}


/*
 * Return the n'th element (from 0) of an array
 */
hn_any *hn_array_arg(n, any)
int n;
hn_any *any;
{
	if (any->type != hn_array_type) {
		hn_error("hn_array_arg: non_array passed\n");
		return((hn_any *)0);
	}
	if ((any->u.array.argc <= n) || (n < 0)) {
		hn_error("hn_array_arg: bad argument passed\n");
		return((hn_any *)0);
	} 
	return(any->u.array.argv[n]);
}


/*
 * Return the n'th element of an array of strings
 */
char *hn_text_line(n, any)
int n;
hn_any *any;
{
	hn_any *arg;

	if (arg = hn_array_arg(n, any))
		if (arg->type == hn_string_type)
			return(arg->u.string);
	return((char *)NULL);	
}



hn_any *hn_make_text(lines)
char **lines;
{
	hn_any *arg;
	char **tpp, **tp = lines;
	int i, nlines;

	for (nlines = 0, tpp = tp; *tpp; tpp++, nlines++);
	arg = hn_new_textmess(nlines);
	
	for (i = 0; i < nlines; i++ )
		hn_array_arg(0, arg)->u.array.argv[i] =
					hn_new_string(tp[i]);

	return(arg);
}
