/* Copyright (c) The Turing Institute 1989 */

#ifndef lint
static char sccsid[] = "@(#)inter.c 3.2 5/28/90 Tim Niblett (Turing Institute)";
#endif lint



/* Display interface to HyperNeWS */

#include <stdio.h>
#include <strings.h>
#include <hypernews.h>

#include "hnutils.h"
#include "hash.h"
#include "readfile.h"
#include "last.h"

char *current_section = "main";

/*
 *	The named objects on the hyperman stack
 */

struct hn_objreg objects[] = {
	{ hn_Stack,	"Hyperman" },
	{ hn_Card,	"HomeCard" },
	{ hn_Card,	"editcard" },
	{ hn_EditText,	"contents" },
	{ hn_EditText,	"section"  },
	{ hn_EditText,	"title"    },
	{ hn_EditText,	"lines"    },
	{ hn_EditText,	"index"    },
	{ hn_Button,    "go" 	   },
	{ hn_Button,    "PreviousButton" },
	{ hn_Button,    "LastButton"     },
	{ hn_Button,    "UpButton"       },
	{ hn_Button,    "NextButton"     },
	{ hn_Stack,		(char *)NULL }
};

/*
 * The handler procedures
 */

int quit(), goto_section(), goto_next_section(), goto_up_section(),
	goto_last_section(), goto_prev_section(), start();

/*
 * The handler message registry
 */

void register_messages() {
	hn_register(quit, hn_new_message(
			ANY,
			"Quit", 
			ANY));
	hn_register(goto_section, hn_new_message(
			ANY,
			"WordAction", 
			hn_new_array_body(1)));
	hn_register(start, hn_new_message(
			(hn_any *)0,
			"GoHome", 
			hn_new_array_body(0)));
	hn_register(goto_last_section, hn_new_message(
			HN_LOOK("LastButton"),
			"GoLastSection", 
			hn_new_array_body(0)));
	hn_register(goto_up_section, hn_new_message(
			HN_LOOK("UpButton"),
			"GoUpSection", 
			hn_new_array_body(0)));
	hn_register(goto_next_section, hn_new_message(
			HN_LOOK("NextButton"),
			"GoNextSection", 
			hn_new_array_body(0)));
	hn_register(goto_prev_section, hn_new_message(
			HN_LOOK("PreviousButton"),
			"GoPrevSection", 
			hn_new_array_body(0)));
}

/*
 *	The interface routines
 */

/*
 *	Set the appropriate text and ancillary info
 *	for the selected section
 */
void set_section_info(sec)
char *sec;
{
	Document *doc;
	register int i;
	hn_any *contents, *file, *caption;
	char lines[20],				/* Only need enough for %d */
		**tp;
	char *long_title();

	if (!(doc = doc_find(sec))) {
		return;
	}
	if (!strncmp("Figure", sec, 6)) { /* We have a figure */
		char *p = *(TEXT(doc)+1), *q;
		q = p + strlen(p);
		while ((q > p) && (*q != '/'))
			q--;
		if (*q == '/')
			q++;

		if (!(file = hn_new_string(q))) {
			hn_error("set_section_info: can't alloc\n");
			return;
		}
		if (!(caption = hn_new_string(TITLE(doc)))) {
			hn_error("set_section_info: can't alloc\n");
			return;
		}
		contents = hn_new_array(caption, file, NULL);
		hn_send(HN_LOOK("Hyperman"), "SetupDrawing", contents);
	} else {
		/* Count the lines in TEXT(doc) */
		for (i = 0, tp = TEXT(doc); *tp; i++, tp++);	

		(void)hn_set_text_line(HN_LOOK("lines"), sprintf(lines, "%d lines", i));	

		contents = hn_make_text(TEXT(doc));
		(void)hn_set_text_line(HN_LOOK("section"), long_title(doc));
		
		hn_send(HN_LOOK("contents"), "SetValue", contents);
	
		current_section = LABEL(doc);
	}

}

int goto_section(any, argc, argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	set_section_info(argv[0]->u.string);
	addlast(argv[0]->u.string);
	return(0);
}

int start(any, argc, argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	char *this_section;
	hn_any *arg;
	void set_up_index();

	set_up_index();

	/* Set the title of the Manual */
	(void)hn_set_text_line(HN_LOOK("title"), TITLE(root));

	set_section_info("main");

	arg = hn_new_array_body(1);
	arg->u.array.argv[0] = hn_new_name("EditCard");
	hn_send(HN_LOOK("Hyperman"), "GotoCard", arg);

	return(0);
}

int goto_next_section(any, argc, argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	char *this_section;

	this_section = next_section(current_section);

	set_section_info(this_section);
	addlast(this_section);
	return(0);
}

int goto_prev_section(any, argc, argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	char *this_section;

	this_section = prev_section(current_section);

	set_section_info(this_section);
	addlast(this_section);
	return(0);
}

int goto_up_section(any, argc, argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	char *this_section;

	this_section = up_section(current_section);

	set_section_info(this_section);
	addlast(this_section);
	return(0);
}

int goto_last_section(any, argc, argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	char *this_section;

	this_section = removelast();
	set_section_info(this_section);
	return(0);

}

int quit()
{
	exit(0);
}

void set_up_index()
{
	Document *doc;
	hn_any *contents;

	if (!(doc = doc_find("index"))) {
		contents = hn_make_text((char **)0);
	} else {
		contents = hn_make_text(TEXT(doc));
	}
	hn_send(HN_LOOK("index"), "SetValue", contents);
}

#define	MAX_TITLE_LENGTH 80

char *long_title(doc)
Document *doc;
{
	static char *tmp = (char *)0;
	
	if (tmp)
		free(tmp);
	tmp = (char *)malloc((unsigned)(strlen(TITLE(doc)) + strlen(LABEL(doc))
				+ 3));

	if (!strcmp("main",LABEL(doc)))
		strcpy(tmp,"");
	else {
		strcpy(tmp, LABEL(doc));
		strcat(tmp,". ");
	}
	strcat(tmp, TITLE(doc));
	return(tmp);
}

main(argc, argv)
int argc;
char *argv[];
{
	if (argc != 2) {
		fprintf(stderr,"usage: %s filename\n",argv[0]);
		exit(1);
	}

	read_in_doc(argv[1]);

	if (!hn_start(argv[0]))
		exit(1);
	
	object_registration("Hyperman", objects);
	register_messages();
	hn_register(quit, hn_quit);

	hn_connect(HN_LOOK("Hyperman"));

	/* Start */
	hn_listen(hn_forever);

	/* Stop */
	hn_stop();
}

