
#ifndef lint
static char sccsid[] = "@(#)readfile.c 3.1 3/19/90 Tim Niblett (Turing Institute)";
#endif lint

/* 
 * Read in a file in HN format to a tree structure with all cross referencing in place.
 */

#include "mactype.h"

#include <stdio.h>
#include <ctype.h>
#ifdef SUN
#include <malloc.h>
#endif

#ifdef MAC
#include <strings.h>
#include <unix.h>
#include <storage.h>
#endif

#include "readfile.h"
#include "rk.h"

#define TRUE		1
#define FALSE		0
#define BUFFERSIZE	30000L

#define ESCCHAR		'{'

typedef unsigned short BOOL;

/*
 *	Token types
 */
#define START_SEC	0
#define SEC		1
#define EOFTYPE		2

typedef struct Token {
	int type;
	char * contents;
} Token;

Token TheTok;

Document *root;

char *		buffer;
char *		bp;
long int	bpos = 0;

/*
 *	Forward declarations
 */

char **convert_to_lines();
void print_lines();
Document *doc_hash();

/*	*******************
 *	* End of Preamble *
 *	*******************
 */


Document *new_doc(parent)
Document *parent;
{
	Document *tmp;

	tmp = (Document *)ALLOC(sizeof(Document));
	tmp -> label = "";
	tmp -> contents = (char **)0;
	tmp -> child = (Document *)0;
	tmp -> parent = parent;
	tmp -> sib = (Document *)0;
	tmp -> presib = (Document *)0;
	return(tmp);
}

Document *set_doc_label(doc, lab)
Document *doc;
char * lab;
{
	char *label = (char *)malloc(strlen(lab)+1);
	char *fromlabel;
	char *strcpy();

	strcpy(label,lab);
	
	if (fromlabel = search("!", label)) {
		/* A bit hacky: we split label into 2 strings
		   by deleting the "!" and replacing with "" */
		TITLE(doc) = fromlabel+1;
		*fromlabel = '\0';
	}
	LABEL(doc) = label;
	/*fprintf(stderr,"label: %s -- %s\n",LABEL(doc),TITLE(doc));*/
	return(DOCSET(doc));
}

Document *set_doc_contents(doc, contents)
Document *doc;
char * contents;
{
	TEXT(doc) = convert_to_lines(contents);
	return(doc);
}

/* 
 *	Buffer Operations
 */

void clear_buffer()
{
	bp = buffer;
	buffer[0] = '\0';
	bpos = 0;
}


/*	**********
 *	Tokenizing
 *	**********
 */


/*
 *	States for the tokenizer
 */

#define PRE_START	0
#define	IN_SEC		1

static int token_state = PRE_START;

/*
 *	advance returns the next char in the input stream, and checks that you
 *	aren't overflowing the buffer.  If the character is escaped it sets
 *	the escape flag to be true.
 */
#define TABWIDTH	8

static BOOL escape = FALSE;
static int charpos = 0;
int advance(f)
FILE *f;
{
	int n;

	escape = FALSE;
	if ((*bp = getc(f)) == EOF) {
		*bp = NULL;
		return(EOF);
	}
	if (*bp == '\\') {
		*bp = getc(f);
		escape = TRUE;
	}
	/* fprintf(stderr, "%c", *bp); */
	if (bpos > BUFFERSIZE-3) {
		fprintf(stderr, "Fatal error: buffer too small\n");
		exit(1);
	}
	switch (*bp) {
	case '\n':
		charpos = 0;
		bpos++;
		return(*bp++);
	case '\t':
		/* add some spaces to mimic tab character */
		for (n = 8 - (charpos % TABWIDTH) ; n ; n--) {
			if (bpos < (BUFFERSIZE -3)) {
				bpos++;
				*bp++ = ' ';
			}
		}
		return ('\t');
	default:
		charpos++;
		bpos++;
		return(*bp++);
	}
}

/*
 *	Make a frest copy of a string.
 */
char * make_copy(string)
char *string;
{
	char *strcpy();

	char * tmp = (char *)ALLOC((strlen(string)+1));
	(void)strcpy(tmp,string);
	return(tmp);
}

int get_token(f)
FILE *f;
{

	register int c;
	char * contents = buffer;

	clear_buffer();
	/* Start having read nothing */		
	while ((c = advance(f)) != EOF) {
		if (escape) 
			continue;
		if ((char)c == '{') {
		  switch (token_state) {
		    case PRE_START:
			contents = bp;
			while ((advance(f) != '}') || escape);
			if (bpos < 1) {
				fprintf(stderr, "Fatal error: buffer underflow\n");
				exit(1);
			}
			bp--;
			bpos--;
			*bp = NULL;
			token_state = IN_SEC;
			TheTok.type = START_SEC;
			TheTok.contents =
				contents;	
			return(START_SEC);
			break;
		    case IN_SEC:
			if (advance(f) == '}' && !escape) {
				if (bpos < 2) {
					fprintf(stderr, "Fatal error: buffer underflow 2\n");
					exit(1);
				}
				bp -= 2;
				bpos -=2;
				*bp = NULL;
				token_state = PRE_START;
				TheTok.type = SEC;
				TheTok.contents = 
					make_copy(contents);
				return(SEC);
			}
			break;
		  }
		}
	}
	TheTok.type = EOFTYPE;
	TheTok.contents = NULL;
	return(EOFTYPE);
}


/*
 * Tree construction operations
 */

/* string is something like s-2-2-3,
 * which means second section, second subsection ... etc.
 * We move to that section creating intermediate trees
 * if necessary.
 */
Document *get_to_node(string)
char * string;
{
	Document * doc = root, *sib;
	int index;
	char * s = string;
	char *p, name[128];
	
	if (!strncmp("main", s, 4)) {
		return(set_doc_label(root,string));
	}
		
	if (strncmp("S", s, 1))
		return(set_doc_label(new_doc(doc),string));

	/* skip over "Section " */
	while (*s && (isspace(*s) || isalpha(*s) || (*s < 3))) s++;


	while (*s && *s != '!') {
		while (*s == '.') s++;
		if (*s && !isdigit(*s)) {
			fprintf(stderr, "get_to_node(%s): funny value\n\n");
			exit(1);
		}
		/* Go down a level */
		if (!CHILD(doc))
			CHILD(doc) = new_doc(doc);
		doc = CHILD(doc);
		
		for (index = 0 ;isdigit(*s); s++)
			index = (index*10) + ((int)(*s) - (int)'0');
		for ( index-- ;index > 0; index-- ) {
			if (SIB(doc))
				doc = SIB(doc);
			else {
				sib = new_doc(PARENT(doc));
				PRESIB(sib) = doc;
				doc = SIB(doc) = sib;
			}
		}
		while (*s && (*s < 3)) s++;
	}
	
	for (p = name ; *string ; string++)
		if (*string >= 6)
			*p++ = *string;
	*p = '\0';

	/* Now add in the string */
	return(set_doc_label(doc,name));
}

Document *current_doc;
void do_parse(f)
FILE *f;
{
	int token_type;
	void do_debug_print();
	
	while ((token_type = get_token(f)) != EOFTYPE) {
		switch (token_type) {
		  case SEC:
			TEXT(current_doc) = convert_to_lines(TheTok.contents);
			/* do_debug_print(current_doc); */
			break;
		  case START_SEC:
			if (!(current_doc = get_to_node(make_copy(TheTok.contents)))) {
				(void)get_token(f);
				current_doc = root;
			}
			break;
		}
	}
}

void do_debug_print(doc)
Document *doc;
{
	printf("***\n{%s}\n", LABEL(doc));
	printf("TITLE: %s\n", TITLE(doc));
	print_lines(stdout, TEXT(doc));
	printf("\n{}\n");

}

char *next_section(sec)
char *sec;
{
	Document *doc;

	if (!(doc = doc_find(sec)))
		return(LABEL(root));
	if (CHILD(doc))
		return(LABEL(CHILD(doc)));
	if (SIB(doc))
		return(LABEL(SIB(doc)));
	while (PARENT(doc)) {
		doc = PARENT(doc);
		if (SIB(doc))
			return(LABEL(SIB(doc)));
	} 
	return(LABEL(root));
}

char *up_section(sec)
char *sec;
{
	Document *doc;
	if (!(doc = doc_find(sec)))
		return(LABEL(root));
	if (PARENT(doc))
		return(LABEL(PARENT(doc)));
	return(LABEL(root));
}

char *prev_section(sec)
char *sec;
{
	Document *doc;

	if (!(doc = doc_find(sec)))
		return(LABEL(root));
	if (PRESIB(doc))
		doc = PRESIB(doc);
	else {
		if (PARENT(doc))
			return(LABEL(PARENT(doc)));
	}
	while (CHILD(doc)) {
		/* scan to the last section */
		doc = CHILD(doc);
		while (SIB(doc))
			doc = SIB(doc);
	}
	return (LABEL(doc));
}

void read_in_doc(file) 
char *file;
{
	FILE *f;
	
	doc_hash_init();
	root = new_doc((Document *)0);
	buffer = (char *)ALLOC(BUFFERSIZE);
	
	if ((f = fopen(file, "r")) == NULL) {
		fprintf(stderr, "Can't open file %s\n",file);
		exit(1);
	}
	do_parse(f);
	/* do_debug_print(root); */

}

#ifdef SOLO
MAIN()
{
	read_in_doc();
}

#endif

/*	***************************
 *	* Some ancillary routines *
 *	***************************
 */
 
/*
 *	The no. of lines in a string -- minimum 1.
 */
int count_lines(text)
register char *text;
{
	register int i;
	for (i=1; *text; text++)
		if (*text == '\n')
			i++;
	return(i);
}

/*
 *	Convert a string (text) into a set of strings
 *	on a line basis.  The original string is
 *	copied.
 */
char **convert_to_lines(text)
char *text;
{
	int count = 0;
	char **tmp, *line;
	register char *ptr = text, **tmptr;
	char *strncpy();

	tmptr = tmp = (char **)ALLOC((count_lines(text) + 1) * sizeof(char *));

	while (*text) {
		ptr = text;
		count = 0;
		while (*ptr && *ptr != '\n') {
			count++;
			ptr++;
		}
		line = (char *)ALLOC(count + 1);
		(void)strncpy(line,text,count);
		line[count] = '\0';
		if (ptr)
			text = ptr+1;
		else 
			text = ptr;
		*tmptr++ = line;
	}
	*tmptr = (char *)0;
	return(tmp);
}

/*
 * 	Print out lines to file
 */
void print_lines(f, lines)
FILE *f;
char **lines;
{
	while (*lines) {
		printf("%s\n", *lines++);
	}
}

/*
 *	Hash access to documents
 */


#define HASHSIZE	256

typedef struct hash_struct
{
	Document  *doc;
	struct hash_struct *link;
} hash_struct;

static hash_struct *hashtable[HASHSIZE];

static int hash(string)
register char *string;
{
	register int h = 0;

	while (*string)
		h += *string++;
	return(h % (HASHSIZE-1));
}


Document  *doc_hash(doc, create)
register Document *doc;
int create;
{
	register int h;
	register hash_struct *p;
	register char *string = LABEL(doc);
	
	if (!string)
		return((Document *)0);
	h = hash(string);
	p =  hashtable[h];
	while (p)
		if (p->doc == doc)
			return(p->doc);
		else
			p = p -> link;
	if (create) {
		p = (hash_struct *) ALLOC(sizeof(hash_struct));
		p -> doc = doc;
		p -> link = hashtable[h];
		hashtable[h] = p;
		return(p->doc);
	} 
	else 
		return((Document *)0);
}

/* Strip out multiple spaces */
char *strip_spaces(string)
char *string;
{
	char *tmp,*tmpstart;

	tmpstart = tmp = (char *)malloc((unsigned)(strlen(string)+1));
	
	while (*string) {
		if ((*string == ' ') && (*(string+1) == ' '))
			string++;
		else
			*tmp++ = *string++;
	}
	*tmp = '\0';
	return(tmpstart);
}

Document  *doc_find(string)
register char *string;
{
	register int h;
	register hash_struct *p;
	
	/*fprintf(stderr, "Input string: <%s>\n", string);*/
	string = strip_spaces(string);
	/*fprintf(stderr, "Search string <%s>\n", string);*/

	if (!string)
		return((Document *)0);
	h = hash(string);
	p =  hashtable[h];
	while (p)
		if (!strcmp(LABEL(p->doc), string))
			return(p->doc);
		else
			p = p -> link;
	return((Document *)0);
}


void doc_hash_init()
{
	register int i;

	for(i = 0; i < HASHSIZE ; i++)
		hashtable[i] = 0;
}

