
#ifndef lint
static char sccsid[] = "@(#)hnserver.c 3.1 3/19/90 Arthur van Hoff (Turing Institute)";
#endif lint

/**********************************************************************
 * file: 	hnserver.ps		(c)1989 Turing Institute
 * author: 	Arthur van Hoff
 *
 * Description:
 * 	This is the HyperNeWS server.
 *	The 'hnserver' is a client which is connect to the
 *	system stack at startup. After that it provides small services
 *	to HyperNeWS. The following services are provided at present:
 *
 *	1) Keep track of the systems time and date. (See SysTime etc. in util.ps)
 *	2) Provide a keyboard-beep (see SysBeep in util.ps).
 */

#include <stdio.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/ioctl.h>
#include <sundev/kbio.h>
#include <sundev/kbd.h>

#include "hnserver.h"
#include "hypernews.h"

extern int hn_intro;

int kbd;
hn_any *systemstack;

int	day = -1, year = -1, month = -1, date = -1, hour = -1, min = -1;

char *months[] =
	{"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
char *days[] =
	{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};

/**********************************************************************
 * Called when approx. 1 minite has passed. Check if any of the time
 * paramers have changed. If so, update them in HyperNeWS.
 */

int checktime()
{
        struct tm *tm;
	hn_any *thetime;
	long now;

	now = time(0);
	tm = localtime(&now);

	if (tm->tm_wday != day) {
		day = tm->tm_wday;
		set_system_day(days[day]);
	}
	if (tm->tm_mday != date) {
		date = tm->tm_mday;
		set_system_date(date);
	}
	if (tm->tm_mon != month) {
		month = tm->tm_mon;
		set_system_month(months[month]);
	}
	if (tm->tm_year != year) {
		year = tm->tm_year;
		set_system_year(year + 1900);
	}
	if (tm->tm_hour != hour) {
		hour = tm->tm_hour;
		set_system_hour(hour);
	}
	if (tm->tm_min != min) {
		min = tm->tm_min;
		set_system_min(min);
	}
}

/**********************************************************************
 * Called when HyperNeWS is killed.
 */

int quit()
{
	exit(0);
}

/**********************************************************************
 * Called from hypernews by 'SysBeep' in util.ps
 */

int sysbeep(any)
hn_any *any;
{
	int     on = KBD_CMD_BELL, off = KBD_CMD_NOBELL;
	struct timeb tb;
	long	t;


	if (!kbd)
		return(0);
        if (ioctl(kbd,KIOCCMD,&on) < 0)
                return(0);

	ftime(&tb);
	t = (tb.time * 1000) + (long)tb.millitm + 
		any->u.message.args->u.array.argv[0]->u.number;

        /* silly delay loop */
	do {
		ftime(&tb);
	} while (((tb.time * 1000) + (long)tb.millitm) < t);

        if (ioctl(kbd,KIOCCMD,&off) < 0)
                return(0);

	return(1);
}

/**********************************************************************
 * main program.
 */

main(argc,argv)
int argc;
char *argv[];
{
        if ((kbd = open("/dev/kbd",O_RDONLY)) < 0)
                kbd = 0;

	/* startup */
	hn_intro = 1;
	hn_start(argv[0]);

	/* define the system stack address */
	systemstack = hn_new_object(NULL,hn_Stack,"system");

	/* register routines */
	hn_register(checktime,hn_timeout);
	hn_register(quit,hn_quit);
	hn_register(sysbeep,
		hn_new_message(ANY,"SysBeep",hn_new_array_body(1)));

	/* connect to the system stack */
	hn_connect(systemstack);

	/* go! */
	checktime(NULL);
	hn_listen(10000);

	/* exit */
	hn_stop();
}
