
#ifndef lint
static char sccsid[] = "@(#)child.c 3.3 9/21/90 Arthur van Hoff (Turing Institute)";
#endif lint


#include <stdio.h>
#include <sys/file.h>
#include <sys/fcntl.h>
#include <sgtty.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sundev/kbio.h>
#include <sundev/kbd.h>
#include "default.h"

#define	MAXDELAY	120
#define DEVTTY		"/dev/tty"
#define KBDDEV		"/dev/kbd"

extern int	errno, rows, cols, console;
extern char	*command, **arguments, **environ;

char	*default_args[] = {DEFAULTSHELL,NULL};
char	erase, kill, *gp, *tp;
char	ptcname[] = "/dev/ptyXX";
char	ptsname[] = "/dev/ttyXX";
int	pid, pgrp, Mfd, Sfd, kbd;
int	ld = NTTYDISC, lm = LCRTBS|LCRTERA|LCRTKIL|LCTLECH;
FILE	*child;
struct	sgttyb sg;
struct	tchars tc;
struct	ltchars ltc;

init_child()
{
	int	delay = 2, n, i;

	/* fork the child process */

	while (((pid = vfork()) < 0) && (errno == EAGAIN)) {
		sleep(delay);
		if ((delay <<= 1) > MAXDELAY)
			perror("cannot fork child","");
	}

	if (pid == 0) {
		/* child process */
		n = getdtablesize();
		setpgrp(0,(pid = getpid()));
		dup2(Sfd,0);
		dup2(Sfd,1);
		dup2(Sfd,2);
		for (i = 3 ; i < n ; i++)
			close(i);
		if (ioctl(0,TIOCSPGRP,&pid) == -1)
			perror("TIOCSPGRP","");
		if (arguments == NULL) {
			arguments = default_args;
			if (command == NULL)
				command = DEFAULTSHELL;
			arguments[0] = command;
		}
		execvp(command,arguments);
		perror("execvp",command);
	}
	pgrp = pid;
	close(Sfd);
}

change_tty_size()
{
	struct ttysize	ts;

	ts.ts_lines = rows;
	ts.ts_cols = cols;
	if ((ioctl(Mfd,TIOCSSIZE,&ts)) == -1)
		perror("failed to change size","");
}

init_tty()
{
	int	fd, gotstate = 0;
	char	*gp, *tp; 
	static	char	str[256];
	int	flags;

	if ((fd = open(DEVTTY,(O_RDWR|O_NDELAY),0)) >= 0) {
		if (ioctl(fd,TIOCGETP,&sg) == 0) {
			flags = sg.sg_flags;
			erase = sg.sg_erase;
			kill = sg.sg_kill;
			sg.sg_flags &= ~(ALLDELAY|RAW|CBREAK);
			ioctl(fd,TIOCGETD,&ld);
			ioctl(fd,TIOCLGET,&lm);
			ioctl(fd,TIOCGETC,&tc);
			ioctl(fd,TIOCGLTC,&ltc);
			gotstate = 1;
		} else
			perror("ioctl","TIOCGETP");
		if (ioctl(fd,TIOCNOTTY,0) == -1)
			perror("ioctl","TIOCNOTTY");
		close(fd);
	}

	close(0);
	close(1);

	for (gp = "pqrst" ; *gp ; gp++)
		for (tp = "6789abcdef012345" ; *tp ; tp++) {
			ptcname[8] = *gp;
			ptcname[9] = *tp;
			if ((Mfd = open(ptcname,2)) < 0)
				continue;
			ptsname[8] = *gp;
			ptsname[9] = *tp;
			if ((Sfd = open(ptsname,2)) >= 0)
				goto done;
			close(Mfd);
		}
done:
	if ((Mfd < 0) || (Sfd < 0))
		perror("could not setup tty","");
	if (ioctl(Sfd,TIOCHPCL,0) == -1)
		perror("ioctl","TIOCHPCL");
	if (gotstate) {
		ioctl(Sfd,TIOCGETP,&sg);
		sg.sg_erase = erase;
		sg.sg_kill = kill;
		sg.sg_flags = flags | (CRMOD|ANYP|ECHO);
		ioctl(Sfd,TIOCSETP,&sg);
		ioctl(Sfd,TIOCSETC,&tc);
		ioctl(Sfd,TIOCSLTC,&ltc);
	}
	ioctl(Sfd,TIOCSETD,&ld);
	ioctl(Sfd,TIOCLSET,&lm);
	if (console) {
		if (ioctl(Sfd,TIOCCONS,0) < 0)
			perror("ioctl","TIOCCONS");
	}
	child = fdopen(Mfd,"r");

	sprintf(str,"TERM=%s",DEFAULTTERM);
	putenv(str);
	unsetenv("TERMCAP");

	if ((kbd = open(KBDDEV,O_RDONLY)) < 0)
		perror("could not open:",KBDDEV);

	add_utmp(ptsname);
}

cleanup_tty()
{
	rm_utmp();
}

tty_bell()
{
	float	i;
	int	on = KBD_CMD_BELL, off = KBD_CMD_NOBELL;

	if (ioctl(kbd,KIOCCMD,&on) < 0) 
		perror("ioctl",KBDDEV);

	/* silly delay loop take it out someday!!!! */
	for (i = 0 ; i < 500 ; i++);

	if (ioctl(kbd,KIOCCMD,&off) < 0) 
		perror("ioctl",KBDDEV);
}

unsetenv(name)
char       *name; 
{
    	register    len;
    	register char **ap, **new;

    	len = strlen(name);
    	for (new = ap = environ; *ap; ap++) {
		if (strncmp(*ap, name, len) != 0) {
	    	*new++ = *ap;
		}
    	}
    	if (new < ap) {
		*new = 0;
    	}
}
