
#ifndef lint
static char sccsid[] = "@(#)hnterm.c 3.3 8/10/90 Arthur van Hoff (Turing Institute)";
#endif lint


#include <stdio.h>
#include <sys/time.h>
#include <ctype.h>
#include "hypernews.h"
#include "screen.h"
#include "hnterm.h"
#include "default.h"

#ifndef PSFILE
#define         filenr(fp)      fileno(fp)
#else
#define         filenr(fp)      psio_fileno(fp)
#endif PSFILE

/******************************************************************
 * parameters 
 */

#define MARGIN	10

int	width, height;
int	fontsize = 18, font_ascent = 12, font_descent = 6;
double	font_width = 10.8;
int	rows, cols;
int	console = 0;
int	selection = 0;
int	cursor_vis = 0;
int	cursor_row = -1, cursor_col = -1;
int	persist = 0;
int	hadresize = 1;
int	history = DEFAULTHISTORY;
int	xpos = -1;
int	ypos = -1;
int	his_val = -1, hmin = 0, hmax = 0, hact = 0;
int mousex, mousey;
int startx, starty;
int doubleclick;

extern int row, col, his_top, rows, Mfd, scroll_from, scroll_to;
extern int his_siz, old_row, old_col;

char 	*progname,
	*stackname = "Terminal",
	*command	= NULL,
	**arguments	= NULL;

char strbuf[512];
char buf[BUFSIZ], *in;
int incount = 0;
extern FILE *child;

/******************************************************************
 * postscript objects
 */

hn_any *mystack, *mycanvas, *myslider;

/******************************************************************
 * routines
 */

#define topvpos()	((height - font_ascent) - MARGIN)
#define vpos(row)	(topvpos() - ((row) * fontsize))
#define hpos(col)	(MARGIN + (int)((col) * font_width))
#define posv(y)		min(rows-1,max(0,((height - MARGIN) - y) / fontsize))
#define posh(x,y)	min(screen[y].len,max(0,\
				((x - MARGIN) + (font_width/2))/font_width))

perror(err,arg)
char *err, *arg;
{
	fprintf(stderr,"%s %s\n",err,arg);
	cleanup_tty();
	exit(0);
}

set_font_info()
{
	hn_any *any;

	hn_set_canvas(mycanvas);
	fontinfo();
	while (hn_read(hn_forever) != hn_sync);
	any = hn_read(hn_forever);
	fontsize = any->u.array.argv[0]->u.number;
	font_ascent = any->u.array.argv[1]->u.number;
	font_descent = fontsize - font_ascent;
	font_width = any->u.array.argv[2]->u.number;

	hn_free(any);
}

set_terminal_size(w,h)
int w, h;
{
	int	r, c;

	mousex = mousey = startx = starty = 0;
	set_font_info();

	hadresize = 1;
	width = w;
	height = h;
	r = (height - (MARGIN*2)) / fontsize;
	c = (width - (MARGIN*2)) / font_width;
	if (r < row) {
		resize_screen((row-r),r,c);
	} else
		resize_screen(0,r,c);
	rows = r;
	cols = c;
	scroll_from = 0;
	scroll_to = rows-1;
	if (row >= rows)
		row = rows - 1;
	if (col >= cols)
		col = cols - 1;
	if (his_top) {
		/* history was visible update everything */
		his_top = 0;
		selection = 0;
		MoveTo(0,0);
		old_row = 0;
		old_col = 0;
	} 
	change_tty_size();
}

InvertRange()
{
}

ClearScreen()
{
	begin();
	hide_cursor();
	background();
	fill(MARGIN,MARGIN,width-(MARGIN*2),height-(MARGIN*2));
	end();
}

ScrollLines(n,from,to)
int n, from, to;
{
	int	range = to - (--from);

	if (range <=0)
		return;
	if (range <= abs(n)) {
		ps_gsave();
		background();
		fill(MARGIN,vpos(to)-font_descent,width-(MARGIN*2),
			(range * fontsize));
		ps_grestore();
		return;
	}
	if (!n)
		return;
	if (n < 0) {
		ps_gsave();
		scroll(MARGIN,vpos(to+n) - font_descent,width-(MARGIN*2),
			(range+n)*fontsize,n*fontsize);
		background();
		fill(MARGIN,vpos(from-n) - font_descent,width-(MARGIN*2),
			(-n)*fontsize);
		ps_grestore();
	} else {
		ps_gsave();
		scroll(MARGIN,vpos(to) - font_descent,width-(MARGIN*2),
			(range-n)*fontsize,n*fontsize);
		background();
		fill(MARGIN,vpos(to) - font_descent,width-(MARGIN*2),
			n*fontsize);
		ps_grestore();
	}
}

PrintLine0(row,col,str,len,invert)
int row, col;
char *str;
int len, invert;
{
	register char *s = strbuf;
	register int n = len;

	if (!len)
		return;

	while (n--) *s++ = (*str++ & 0177);
	*s = '\0';

	if (invert) {
		ps_gsave();
		fill(hpos(col),vpos(row)-font_descent,
			(int)(len * font_width),fontsize);

		background();
		drawstr(hpos(col),vpos(row),strbuf,len);
		ps_grestore();
	} else {
		ps_gsave();
		background();
		fill(hpos(col),vpos(row)-font_descent,
			(int)(len * font_width),fontsize);
		ps_grestore();
		drawstr(hpos(col),vpos(row),strbuf,len);
	}
}

EraseLine(row)
int row;
{
	ps_gsave();
	background();
	fill(MARGIN,vpos(row)-font_descent,
		width - (MARGIN*2),fontsize);
	ps_grestore();
}

PutLine(row,str,len)
int row;
char *str;
int len;
{
	int col = 0;
	register char *s = str;


	/* put a line on the screen (remember to invert if needed) */
	/*drawstring(hpos(0),vpos(row),str);*/
	while (*s) {
		while (*s && (!(*s & 0200)))
			s++;
		if (s > str) {
			/* print not inverted */
			drawstr(hpos(col),vpos(row),str,s-str);
			len -= s-str;
			col += s-str;
			str = s;
		}
		while (*s && (*s & 0200))
			s++;
		if (s > str) {
			/* print inverted */
			PrintLine0(row,col,str,s-str,1);
			len -= s-str;
			col += s-str;
			str = s;
		}
	}
}

PrintLine(row,col,str,len)
int row, col;
register char *str;
int len;
{
	register char *s = str;

	/* put part of a line on the screen 
	  (remember to erase/invert if needed) */

	while (*s) {
		while (*s && (!(*s & 0200)))
			s++;
		if (s > str) {
			/* print not inverted */
			PrintLine0(row,col,str,s-str,0);
			len -= s-str;
			col += s-str;
			str = s;
		}
		while (*s && (*s & 0200))
			s++;
		if (s > str) {
			/* print inverted */
			PrintLine0(row,col,str,s-str,1);
			len -= s-str;
			col += s-str;
			str = s;
		}
	}
}

MoveTo(row,col)
{
	if (cursor_vis) {
		DrawCursor();
		cursor_vis = 0;
	}
	cursor_row = row;
	cursor_col = col;
}

DrawCursor()
{
	reverse(hpos(cursor_col),(vpos(cursor_row)-font_descent),
		(int)font_width,fontsize);
}

UndoSelection()
{
}

ScrollValue(val)
int val;
{
	hn_any *any;

	if (val != his_val) {
		his_val = val;
		val = hmin + (hmax - (val - hmin));
		/* do it */
		ps_gsave();
		hn_send(myslider,"SetValue",any = hn_new_integer(val));
		hn_free(any);
		ps_grestore();
	}
}

ScrollRange(minv,val,maxv)
int minv, val, maxv;
{
	hn_any *any;

	if ((hmin == minv) && (hmax == maxv) && (his_val == val))
		return;
	hmin = minv;
	hmax = maxv;
	his_val = val;
	val = hmin + (hmax - (val - hmin));
	ps_gsave();
	hn_send(myslider,"SetRange",
		any = hn_new_array(
			hn_new_integer(hmin),
			hn_new_integer(val),
			hn_new_integer(hmax),
			NULL));
	hn_free(any);
	ps_grestore();
}

yywrap()
{
	/* dummy wrap routine */
	return(1);
}

output(c)
{
}

/******************************************************************
 * event handlers
 */

newsize(any,argc,argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	set_terminal_size((int)argv[0]->u.number,(int)argv[1]->u.number);
	return(1);
}

newfont()
{
	set_terminal_size(width,height);
	update_screen(0,0,width,height);
	hadresize = 0;
	return(1);
}

update(any,argc,argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	hn_set_canvas(mycanvas);

	if (hadresize) {
		update_screen(0,0,width,height);
		hadresize = 0;
		return(1);
	}
	update_screen((int)argv[0]->u.number,(int)argv[1]->u.number,
		      (int)argv[2]->u.number,(int)argv[3]->u.number);
	return(1);
}

slider(any,argc,argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	int val = argv[0]->u.number;

	his_val = hmin + (hmax - (val - hmin));
	history_scroll(his_val);
	return(1);
}

unput(c)
{
	*--in = c;
	incount++;
}

fillbuf()
{
	struct timeval tv;
	int n, mask = (1 << fileno(child)) | (1 << filenr(PostScriptInput));

	tv.tv_sec = 0;
	tv.tv_usec = 100;

	if (n = select(32,&mask,0,0,&tv)) {
		if (mask & (1 << filenr(PostScriptInput)))
			return(0);
		if ((n = read(fileno(child),buf,BUFSIZ)) > 0) {
			incount = n;
			in = buf;
			return(1);
		} else
			return(0);
	} else
		return(0);
}

input()
{
	if (incount) {
		incount--;
		return(*in++);
	} else 
		if (fillbuf()) {
			incount--;
			return(*in++);
		} else
			return(0);
}

pending()
{

	/* input from child */
	if (fillbuf()) {
		yylex();
		flush_output();
	}
	return(1);
}

copy(any)
hn_any *any;
{
	int ln = starty;
	int n,i = 0;
	hn_any *text,*mess;
	char *block,*p;
	int MaxLineLen = 255;

	if (mousex || mousey || startx || starty) {
		block = p = hn_alloc((mousey-starty+1)*MaxLineLen);
		text = hn_new_array_body(mousey-starty+1);
		if (starty == mousey) {
			n = min((mousex - startx),MaxLineLen);
			strncpy(p,&screen[ln].ln[startx],n);
			text->u.array.argv[0] = hn_new_string(p);
			*(p+n+1) = '\0';
		}
		else {
			n = min((screen[ln].len - startx),MaxLineLen);
			strncpy(p,&screen[ln].ln[startx],n);
			text->u.array.argv[i++] = hn_new_string(p);
			p += n + 1; *(p++) = '\0';
			while (++ln < mousey) {
				n = min((screen[ln].len),MaxLineLen);
				strncpy(p,screen[ln].ln,n);
				text->u.array.argv[i++] = hn_new_string(p);
				p += n + 1; *(p++) = '\0';
			}
			n = min(mousex,MaxLineLen);
			strncpy(p,screen[ln].ln,n);
			text->u.array.argv[i] = hn_new_string(p);
			*(p+n+1) = '\0';
		}
		mess = hn_new_array(text,NULL);
		hn_send(mystack,"CopyToClip",mess);
		hn_set_canvas(mycanvas);
		hn_destroy(block);
		hn_free(mess);
	}
	return 1;
}

ioerror()
{
	quit();
}

quit()
{
	hn_hide(mystack);
	hn_stop();
	cleanup_tty();
	exit(0);
}

ascii(any)
hn_any *any;
{
	char str[2];

	if (any->type != hn_number_type)
		return(0);

	if ((row + his_top) >= rows)
		history_reset();
	str[0] = any->u.number;
	write(Mfd,str,1);

	return(1);
}

other(any)
hn_any *any;
{
	fprintf(stderr,"other: ");
	hn_print(stderr,any);
	fprintf(stderr,"\n");
	return(1);
}

/******************************************************************
 * Handling the selection
 */

normal_range(x1,y1,x2,y2)
int *x1, *y1, *x2, *y2;
{
	if ((*y1 > *y2) || ((*y1 == *y2) && (*x1 > *x2))) {
		int t = *x1;
		*x1 = *x2; *x2 = t;
		t = *y1; *y1 = *y2; *y2 = t;
	} 
}

invert_line(row,from,to)
int row, from, to;
{
	if (from < to) {
		reverse(hpos(from),(vpos(row)-font_descent),
			((int)font_width * (to - from)),fontsize);
		ps_flush_PostScript();
	}
}

invert_range(x1,y1,x2,y2)
int x1, y1, x2, y2;
{
	normal_range(&x1,&y1,&x2,&y2);
	if (y1 == y2)
		invert_line(y1,x1,x2);
	else {
		invert_line(y1++,x1,cols);
		while (y1 < y2)
			invert_line(y1++,0,cols);
		invert_line(y2,0,x2);
	}
}

undo_selection()
{
	if (mousex || mousey || startx || starty) {
		invert_range(mousex,mousey,startx,starty);
		mousex = mousey = startx = starty = 0;
	}
}

draw_selection()
{
	invert_range(mousex,mousey,startx,starty);
}

begin_word(x,y)
int *x, *y;
{
	int i = min(*x,screen[*y].len-1);
	for (; i > 0 ; i--)
		if (isspace(screen[*y].ln[i])) {
			i++; 
			break;
		}
	*x = max(0,min(i,*x));
}

end_word(x,y)
int *x, *y;
{
	int i = max(0,*x);
	for (; i < screen[*y].len ; i++)
		if (isspace(screen[*y].ln[i]))
			break;
	*x = min(screen[*y].len,max(i,*x));
}

mouse_double(any,argc,argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	int x = argv[0]->u.number;
	int y = argv[1]->u.number;

	undo_selection();
	doubleclick = 1;
	y = posv(y);
	x = posh(x,y);
	startx = mousex = x;
	starty = mousey = y;
	begin_word(&startx,&starty);
	end_word(&mousex,&mousey);
	invert_range(mousex,mousey,startx,starty);
	return 1;
}

mouse_down(any,argc,argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	int x = argv[0]->u.number;
	int y = argv[1]->u.number;

	undo_selection();
	doubleclick = 0;
	y = posv(y);
	x = posh(x,y);
	startx = mousex = x;
	starty = mousey = y;
	return 1;
}

mouse_drag(any,argc,argv)
hn_any *any;
int argc;
hn_any *argv[];
{
	int x = argv[0]->u.number;
	int y = argv[1]->u.number;
	y = posv(y);
	x = posh(x,y);
	if (doubleclick) {
		if ((y < starty) || ((y == starty) && (x < startx)))
			begin_word(&x,&y);
		else
			end_word(&x,&y);
	}
	invert_range(mousex,mousey,x,y);
	mousex = x;
	mousey = y;
	return 1;
}

mouse_up(any)
hn_any *any;
{
	normal_range(&startx,&starty,&mousex,&mousey);
	return 1;
}

/******************************************************************
 * main
 */

main(argc,argv)
int argc;
char *argv[];
{
	char *getenv();
	hn_any *any;

	progname = argv[0];

	if ((argc > 1) && (!strcmp(argv[1],"-c"))) {
		stackname = "Console";
		console = 1;
		argc--;
		argv++;
	}
	if (argc == 2) {
		stackname = argv[1];
	} else
		if (argc != 1) {
			fprintf(stderr,"usage: %s [-c] stackname\n",progname);
			exit(0);
		}

	hn_start(progname);
	hn_verbose = 0;

	mystack = hn_new_object(NULL,hn_Stack,stackname);
	hn_constant(mystack);
	if (!hn_exists(mystack)) {
		fprintf(stderr,"no such stack:");
		hn_print(stderr,mystack);
		fprintf(stderr,"\n");
		hn_stop();
		exit(0);
	}
	mycanvas = hn_new_object(stackname,hn_Canvas,"DataCanvas");
	hn_constant(mycanvas);
	myslider = hn_new_object(stackname,hn_Slider,"DataSlider");
	hn_constant(myslider);

	hn_connect(mystack);
	/* hn_show(mystack); */

	hn_register(newsize,
		hn_new_message(ANY,"NewSize",hn_new_array_body(2)));
	hn_register(update,
		hn_new_message(ANY,"OnUpdate",hn_new_array_body(4)));
	hn_register(newfont,
		hn_new_message(ANY,"NewFont",ANY));
	hn_register(copy,
		hn_new_message(ANY,"TermCopy",ANY));
	hn_register(mouse_down,
		hn_new_message(ANY,"MouseDown",hn_new_array_body(2)));
	hn_register(mouse_double,
		hn_new_message(ANY,"MouseDouble",hn_new_array_body(2)));
	hn_register(mouse_drag,
		hn_new_message(ANY,"MouseDrag",hn_new_array_body(2)));
	hn_register(mouse_up,
		hn_new_message(ANY,"MouseUp",hn_new_array_body(2)));
	hn_register(slider,
		hn_new_message(myslider,"Action",hn_new_array_body(1)));
	hn_register(quit,
		hn_new_message(ANY,"CloseStack",ANY));
	hn_register(pending,hn_pending);
	hn_register(ioerror,hn_ioerror);
	hn_register(quit,hn_quit);
	hn_register(ascii,ANY);
	hn_register(other,ANY);

	hn_set_canvas(mycanvas);
	hn_write(mycanvas);
	prepare_canvas();

	/* Set up command to be the right shell (TN) */
	if (!(command = getenv("SHELL")))
		command = DEFAULTSHELL;

	init_signal();
	init_tty();
	init_child();

	/* get the size of the terminal canvas */
	any = hn_get(mycanvas,"Width");
	width = any->u.number;
	any = hn_get(mycanvas,"Height");
	height = any->u.number;
	set_terminal_size(width,height);

	if (history)
		alloc_history(history);
	ScrollRange(0,0,0);

	/* always listen to your children */
	hn_mask = 1 << fileno(child);

	hn_listen(hn_forever);
	hn_stop();
	cleanup_tty();
}
