
#ifndef lint
static char sccsid[] = "@(#)screen.c 3.1 3/19/90 Arthur van Hoff (Turing Institute)";
#endif lint


#include <stdio.h>
#include "screen.h"
#include "hnterm.h"

extern	int his_top, his_siz, selection;
screen_line	*screen = NULL;
char	*lines = NULL;
int	upd_scroll = 0;

resize_screen(off,r,c)
{
	screen_line	*newscreen;
	char	*newlines;
	int	i;

	if (screen == NULL) {
		screen = (screen_line *)malloc(r * sizeof(screen_line));
		lines = (char *)malloc(r * (c+1));
		for (i = 0 ; i < r ; i++) {
			screen[i].ln = &lines[i*(c+1)];
			screen[i].from = screen[i].to = 0;
			screen[i].len = screen[i].old_len = 0;
			lines[i*(c+1)] = '\0';
		}
	} else {
		if ((r == rows) && (c == cols)) 
			return;
		if ((c <= cols) && (r <= rows)) {
			for (i = 0 ; i < rows ; i++) {
				screen[i].ln[c] = '\0';
				screen[i].len = min(screen[i].len,c);
				screen[i].old_len = min(screen[i].old_len,c);
				screen[i].from = min(screen[i].from,c);
				screen[i].to = min(screen[i].to,c);
			}
			return;
		} 
		if (r > rows)
			newscreen = (screen_line *)
				malloc(r * sizeof(screen_line));
		else
			newscreen = screen;

		newlines = (char *)malloc(r * (c+1));
		for (i = 0 ; i < r ; i++) {
			newscreen[i].ln = &newlines[i*(c+1)];
			if (i < rows) {
				strcpy(&newlines[i*(c+1)],screen[i+off]);
				newscreen[i].from = screen[i+off].from;
				newscreen[i].to = screen[i+off].to;
				newscreen[i].len = screen[i+off].len;
				newscreen[i].old_len = screen[i+off].old_len;
			} else {
				newlines[i*(c+1)] = '\0';
				newscreen[i].from = 0;
				newscreen[i].to = 0;
				newscreen[i].len = 0;
				newscreen[i].old_len = 0;
			}
		}

		free(lines);
		lines = newlines;
		if (r > rows) {
			free(screen);
			screen = newscreen;
		}
	}
}

scroll_screen(dir,from,to)
{
	char 	*p;
	int	i;

	if (his_top)
		history_reset();

	if ((upd_scroll > rows) ||
	    ((dir < 0) && (upd_scroll > 0)) ||
	    ((dir > 0) && (upd_scroll < 0)))
		flush_screen();
	upd_scroll += dir;

	if (dir > 0) {
		if ((from == 0) && (to == (rows-1)))
			add_history(screen[0].ln);
		p = screen[from].ln;
		for (i = from ; i < to ; i++) {
			screen[i].ln = screen[i+1].ln;
			screen[i].from = screen[i+1].from;
			screen[i].to = screen[i+1].to;
			screen[i].len = screen[i+1].len;
			screen[i].old_len = screen[i+1].old_len;
		}
		screen[to].ln = p;
		screen[to].from = 0;
		screen[to].to = 0;
		screen[to].len = 0;
		screen[to].old_len = 0;
	} else {
		p = screen[to].ln;
		for (i = to ; i > from ; i--) {
			screen[i].ln = screen[i-1].ln;
			screen[i].from = screen[i-1].from;
			screen[i].to = screen[i-1].to;
			screen[i].len = screen[i-1].len;
			screen[i].old_len = screen[i-1].old_len;
		}
		screen[from].ln = p;
		screen[from].from = 0;
		screen[from].to = 0;
		screen[from].len = 0;
		screen[from].old_len = 0;
	}
	*p = '\0';
}

clear_screen(from,to)
{
	int	i;

	history_reset();
	for (i = from ; i <= to ; i++) {
		screen[i].ln[0] = '\0';
		screen[i].from = 0;
		screen[i].to = 0;
		screen[i].len = 0;
		screen[i].old_len = 0;
	}
}

reprint_line(row,len,ln)
char *ln;
{
	hide_cursor();
	if (selection) {
		UndoSelection();
		selection = 0;
	}
	EraseLine(row);
	PutLine(row,ln,len);
}

reprint_part(row,from,to,ln)
char *ln;
{
	hide_cursor();
	if (selection) {
		UndoSelection();
		selection = 0;
	}
	PrintLine(row,from,ln+from,to-from);
}

update_screen(x,y,w,h)
int x, y, w, h;
{
	register int	i;

	begin();
	clip(x,y,w,h);
	hide_cursor();
	ClearScreen();
	for (i = 0 ; i < rows ; i++)
		if ((his_top+i) < rows) {
			reprint_line(his_top+i,screen[i].len,screen[i].ln);
			screen[i].from = screen[i].to = 0;
			screen[i].old_len = screen[i].len;
		}
	show_cursor();
	end();
}

flush_screen()
{
	register int	i;

	begin();
	undo_selection();
	/* first check scrolling bits */
	if (upd_scroll) {
		hide_cursor();
		if (selection) {
			UndoSelection();
			selection = 0;
		}
		ScrollLines(upd_scroll,scroll_from,scroll_to);
		upd_scroll = 0;
	}

	/* check if there is something to update */
	for (i = 0 ; i < rows ; i++)
		if ((his_top+i) >= rows) {
			/* don't update lines which are not on the screen! */
			screen[i].from = screen[i].to = 0;
			screen[i].old_len = screen[i].len;
		} else if (screen[i].old_len > screen[i].len) {
			/* line has become shorter */
			reprint_line(his_top+i,screen[i].len,screen[i].ln);
			screen[i].from = screen[i].to = 0;
			screen[i].old_len = screen[i].len;
		} else if (screen[i].from < screen[i].to) {
			/* the contents of the line has changed */
			if (((screen[i].to - screen[i].from) > 5) ||
			    (!screen[i].from && 
			     (screen[i].to == screen[i].len))) {
				/* reprint the whole line */
				reprint_line(his_top+i,screen[i].len,
						screen[i].ln);
				screen[i].from = screen[i].to = 0;
				screen[i].old_len = screen[i].len;
			} else {
				/* reprint a part of the line */
				reprint_part(his_top+i,screen[i].from,
					screen[i].to,screen[i].ln);
				screen[i].from = screen[i].to = 0;
				screen[i].old_len = screen[i].len;
			}
		}
	end();
}

screen_put(c,row,col)
{
	int	i;

	/* put a character at (row,col) on screen */
	/* change the data and update info accordingly */

	if (!(c & 0177)) {
		/* new end of line!! */
		if (col == screen[row].len)
			return;
		for (i = screen[row].len ; i < col ; i++)
			screen[row].ln[i] = ' ';
		if (screen[row].len < col) {
			if (screen[row].from == screen[row].to)
				screen[row].from = screen[row].len;
			screen[row].to = col;
		} else if (screen[row].from != screen[row].to) {
			screen[row].from = min(screen[row].from,col);
			screen[row].to = min(screen[row].to,col);
		}
		screen[row].len = col;
	} else {
		if (screen[row].ln[col] == c)
			return;
		if (col >= screen[row].len)
			screen_put('\0',row,col+1);
		if (screen[row].from == screen[row].to) {
			screen[row].from = col;
			screen[row].to = col+1;
		} else {
			screen[row].from = min(screen[row].from,col);
			screen[row].to = max(screen[row].to,col+1);
		}
	}
	screen[row].ln[col] = c;
}

screen_ensure_pos(row,col)
{
	if (screen[row].len < col)
		screen_put('\0',row,col);
}

refresh_screen()
{
	int	i;
	/* reprint all lines (called after resize when history visible) */
	
	undo_selection();
	upd_scroll = 0;
	for (i = 0 ; i < rows ; i++) {
		screen[i].from = screen[i].to = 0;
		screen[i].old_len = screen[i].len;
		reprint_line(i,screen[i].len,screen[i].ln);
	}
}

screen_ic(r,c)
{
	int	i = screen[r].len;
	/* insert a space at (r,c) */
	if ((screen[r].from > c) || (screen[r].from == screen[r].to))
		screen[r].from = c;
	while (i >= c) {
		screen[r].ln[i+1] = screen[r].ln[i];
		i--;
	}
	screen[r].ln[c] = ' ';
	screen[r].len++;
	screen[r].to = screen[r].len;
}

screen_dc(r,c)
{
	/* delete the character at (r,c) */
	if ((screen[r].from > c) || (screen[r].from == screen[r].to))
		screen[r].from = c;
	while (c < screen[r].len) {
		screen[r].ln[c] = screen[r].ln[c+1];
		c++;
	}
	screen[r].len--;
	screen[r].to = screen[r].len;
}
