
#ifndef lint
static char sccsid[] = "@(#)perfmon.c 3.1 4/5/90 Arthur van Hoff (Turing Institute)";
#endif lint

#include	<stdio.h>
#include	<utmp.h>
#include	<rpcsvc/rstat.h>
#include	<rpcsvc/rusers.h>
#include	<sys/dk.h>
#include	<sys/time.h>
#include	<hypernews.h>

#define		MAXLOAD	1000
#define		MAXU	300

#ifndef PSFILE
#define		filenr(fp)	fileno(fp)
#else
#define		filenr(fp)	psio_fileno(fp)
#endif PSFILE

struct	timeval	tm;
struct	statstime rst;
struct	utmpidlearr up;
struct	utmpidle *uia[MAXU];
char	host[128], *prog, name[10], uhost[20];
int	oldtime[CPUSTATES];
int	i, t, delay, nusers = 0, hadbefore, c = 0;
hn_any	*stack, *title, *button, *arg0, *arg2, *arg4;
FILE	*fp;


void debug(s)
char *s;
{
	fprintf(fp,"debug: %s\n",s);
}


/*
 * quit(): Registered routine to exit the perfmon client process
 */
int quit()
{
	exit(0);
}


/*
 * SendUsers(): Send strings of current users to the perfmon stack
 */
void SendUsers()
{
	up.uia_arr = uia;
	up.uia_cnt = MAXU;
	if (rusers(host,&up)) {
		fprintf(fp,"perfmon: could not get rusers\n");
		exit(-1);
	}
	nusers = up.uia_cnt;
	hn_send(button,"ClearUsers",arg0);
	for (i = 0 ; i < nusers ; i++) {
		hadbefore = 0;
		for (t = 0 ; t < i ; t++) {
			if (!strcmp(
			    up.uia_arr[i]->ui_utmp.ut_name,
			    up.uia_arr[t]->ui_utmp.ut_name)) {
				hadbefore = 1;
				break;
			}
		}
		for (t = 0 ; t < 8 ; t++) {
			name[t] = 
			   up.uia_arr[i]->ui_utmp.ut_name[t];
			uhost[t] = 
			   up.uia_arr[i]->ui_utmp.ut_host[t];
		}
		name[8] = '\0';
		uhost[8] = '\0';
		if (!hadbefore) {
			arg2->u.array.argv[0] = hn_new_string(name);
			arg2->u.array.argv[1] = hn_new_string(uhost);
			hn_send(button,"AddUser",arg2);
		}
	}
	hn_send(button,"ShowUsers",arg0);
}


/*
 * SendStats(): Registered routine to send system load info
 *	to the Perfmon stack
 */
int SendStats()
{
	if (rstat(host,&rst) != 0) {
		fprintf(fp,"perfmon: could not rstat\n");
		exit(-1);
	}
	for (i = 0 ; i < CPUSTATES ; i++) {
		t = rst.cp_time[i];
		rst.cp_time[i] -= oldtime[i];
		oldtime[i] = t;
	}
	for (t = 0, i = 0 ; i < CP_IDLE ; i++)
		t += rst.cp_time[i];
	if (t > 0) {
		arg4->u.array.argv[0] = hn_new_number((double)
			(100*rst.avenrun[0])/MAXLOAD);
		arg4->u.array.argv[1] = hn_new_number((double)
			(100*rst.cp_time[CP_SYS]) / (rst.cp_time[CP_IDLE] + t));
		arg4->u.array.argv[2] = hn_new_number((double)
			(100*rst.cp_time[CP_USER]) / (rst.cp_time[CP_IDLE] + t));
		arg4->u.array.argv[3] = hn_new_number((double)
			(100*rst.cp_time[CP_NICE]) / (rst.cp_time[CP_IDLE] + t));
		hn_send(button,"AddStats",arg4);
	}
	if (nusers != rnusers(host))
		SendUsers();

	return(1);
}


/*
 * main(): Main routine that drives the perfmon stack
 */
int main(argc,argv)
int	argc;
char	*argv[];
{
	fp = fopen("/dev/console","w");
        if (!hn_start(argv[0])) {
		fprintf(fp,"perfmon: error in hn_start()\n");
		exit(1);
	}

	gethostname(host,128);
	delay = 5;

	prog = *argv++;
	while (--argc > 0) {
		if (!strcmp(*argv,"-d")) {
			argv++;
			argc--;
			delay = atoi(*argv++);
		} else if (!strcmp(*argv,"-h")) {
			argv++;
			argc--;
			strcpy(host,*argv);
		} else {
			fprintf(fp,"usage: perfmon [-d delay][-h host]\n");
			exit(-1);
		}
	}

	/* Register routines */
	hn_register(quit,hn_quit);
	hn_register(SendStats,hn_timeout);

	/* Define stack objects we'll reference */
	stack = hn_new_object((char *) NULL, hn_Stack, "Perfmon");
	title = hn_new_object("Perfmon", hn_EditText, "Title");
	button = hn_new_object("Perfmon", hn_Canvas, "PerfButton");
	arg0 = hn_new_array_body(0);
	arg2 = hn_new_array_body(2);
	arg4 = hn_new_array_body(4);

	/* Connect to the system stack */
	if (!hn_connect(stack)) {
		fprintf(fp,"perfmon: error in hn_connect()\n");
		exit(1);
	}

	/* Change the title of the stack to the host name */
	hn_send(title,"SetValue",hn_new_array(hn_new_string(host),NULL));

	if (rstat(host,&rst) != 0) {
		fprintf(fp,"perfmon: could not rstat\n");
		exit(-1);
	}
	for (i = 0 ; i < CPUSTATES ; i++)
		oldtime[i] = rst.cp_time[i];

	/* Start looping & listening */
	SendStats();
	/* Set client to timeout & call SendStats() every "delay" seconds */
	hn_listen(delay * 1000);
	/* This code will never be executed.  Why? */
	hn_stop();
}
