
/* "@(#)hypernews.h 3.2 9/21/90 Arthur van Hoff (Turing Institute)"; */

/****************************************************************************
file: 	hypernews.h		(c)1989 Turing Institute
author:	Arthur van Hoff

Description:
	HyperNeWS interface definitions for HyperNeWS client programs.
*/

/****************************************************************************
 *constant definitions
 */

#define ANY			NULL
#define hn_nr_object_types	9
#define hn_forever		-1

/****************************************************************************
 * HyperNeWS object types, These numbers should match the ones
 * declared in the file $HNHOME/etc/HyperNeWS/client.ps
 */

typedef enum {
	hn_Stack = 1,
	hn_BackGround,
	hn_Card,
	hn_Button,
	hn_EditText,
	hn_Slider,
	hn_PullDown,
	hn_ColorSelect,
	hn_Canvas
} hn_type;

/****************************************************************************
 * hn_any types (see hn_any below)
 */

typedef enum {
	hn_number_type,
	hn_boolean_type,
	hn_string_type,
	hn_name_type,
	hn_array_type,
	hn_null_type,
	hn_message_type,
	hn_object_type
} hn_any_type;


/****************************************************************************
 * The hn_any data structure.
 * This data structure is used to hold any HyperNeWS data send/received
 * to/from HyperNeWS.
 */

typedef struct hn_any {
	hn_any_type	type;
	int		refcount;
	union	{
		double			number;
		int			boolean;
		char			*string;	/* null terminated */
		char			*name;		/* null terminated */
		struct {
			struct hn_any	*target;	/* object */
			char		*name;		/* string */
			struct hn_any	*args;		/* array of arguments */
		} message;
		struct {
			int		argc;
			struct hn_any	**argv;		/* array of hn_any */
		} array;
		struct {
			char		*stack;		/* stack name (or null) */
			hn_type		type;		/* object type nr */
			char		*name;		/* object name */
		} object;
		/* nulltype has no value */
	} u;
} hn_any;

/****************************************************************************
 * External variables
 */

extern int 	hn_status;
extern int 	hn_running;
extern int 	hn_verbose;
extern int 	hn_mask;;
extern char 	*hn_object_type_names[];
extern hn_any 	*hn_true, 
		*hn_false, 
		*hn_null, 
		*hn_ioerror, 
		*hn_timeout,
		*hn_quit,
		*hn_nullarray,
		*hn_sync,
		*hn_pending;


/****************************************************************************
 * Useful things.
 */

/* is true when all seems to be ok */
#define hn_ok()			(hn_running && hn_status)

/* use an hn_any data structure */
#define hn_use(any)		((any)->refcount++)

/* make an hn_any datastructure constant (it will never be deleted) */
#define hn_constant(any)	((any)->refcount = 128)

/* Check if an object is valid */
#define hn_valid_object(obj)	\
		((obj) && ((obj)->type == hn_object_type) &&\
		 ((obj)->u.object.name) &&\
		 (((obj)->u.object.type == hn_Stack) ||\
		  ((obj)->u.object.stack)))


/****************************************************************************
 * External routines.
 */

/* main.c */
extern int hn_start();
extern int hn_stop();
extern void hn_register();
extern int hn_listen();

/* util.c */
extern void *hn_alloc();
extern void hn_destroy();
extern void hn_free();
extern char *str2name();
extern hn_any *hn_new_number();
extern hn_any *hn_new_string();
extern hn_any *hn_new_name();
extern hn_any *hn_new_object();
extern hn_any *hn_new_array_body();
extern hn_any *hn_new_array();
extern hn_any *hn_new_message();
extern int hn_match();
extern void hn_print();

/* io.c */
extern void hn_flush();
extern void hn_flush_input();
extern hn_any *hn_read();
extern hn_any *hn_read_sync();
extern int hn_write();

/* intf.c */
extern hn_any *hn_get();
extern int hn_put();
extern int hn_send();
extern int hn_show();
extern int hn_hide();
extern int hn_exists();
extern int hn_connect();
extern int hn_set_canvas();
extern int hn_set_drawing();
extern hn_any *hn_stack_copy();
extern hn_any *hn_rename();
extern int hn_show_error();
extern int hn_ps();

/****************************************************************************
 * Virtual routines.
 */

#define hn_new_null()		hn_null
#define hn_new_boolean(val)	((val) ? hn_true : hn_false)
#define hn_new_integer(i)	hn_new_number((double)(i))

#define hn_setvalue(object,val)	hn_send(object,"SetValue",val)
#define hn_lock(object)		hn_send(object,"LockUpdates",NULL)
#define hn_unlock(object)	hn_send(object,"UnLockUpdates",NULL)
#define hn_draw(object)		hn_send(object,"Draw",NULL)
#define hn_update(object)	hn_send(object,"Update",NULL)

