
#ifndef lint
static char sccsid[] = "@(#)intf.c 3.2 9/21/90 Arthur van Hoff (Turing Institute)";
#endif lint

/****************************************************************************
file: 	intf.c		(c)1989 Turing Institute
author:	Arthur van Hoff

Description:
	Interface routines for HyperNeWS client programs.

	hn_any *hn_get(hn_any *object, char *param)
		get a parameter value of an object.
	int hn_put(hn_any *object, char *param, hn_any *value)
		set a parameter value of an object.

	int hn_exists(hn_any *object)
		Returns 1 if the object exists, 0 otherwise.

	int hn_send(hn_any *target,char *message, hn_any *args)
		Sends the message with args to the target.
		Returns 0 when the connection with HyperNeWS is lost,
		0 otherwise.

	int hn_show(hn_any *object)
		Show an object.  Returns 0 when the connection with 
		HyperNeWS is lost, 0 otherwise.

	int hn_hide(hn_any *object)
		Hide an object.  Returns 0 when the connection with 
		HyperNeWS is lost, 0 otherwise.

	int hn_connect(hn_any *object)
		Connect the client to a stack. All event from the stack
		will then be send to this client.  Returns 0 when the 
		connection with HyperNeWS is lost, 0 otherwise.

	int hn_set_canvas(hn_any *object)
		Set the client's canvas to the canvas of the indicated
		canvas object. The font is also set.

	int hn_set_drawing(hn_any *object, char *file)
		Put a drawing from a file into a button.
	
	hn_any *hn_rename(hn_any *object, char *name)
		Rename an object (not a stacks). Returns the resulting object.

	hn_any *hn_stack_copy(hn_any *object)
		Create a copy of the named stack. The new stack object is
		returned.

	int hn_show_error(hn_any *text)
		Show an error message using the 'ErrorMessage' stack.

	int hn_ps(char *str)
		Send a postscript command to the HyperNeWS server.
		Use at your own risk.  Returns 0 when the connection with 
		HyperNeWS is lost, 0 otherwise.

*/

#include <stdio.h>
#include "hnps.h"
#include "token.h"
#include "hypernews.h"

#define RESULTDELAY	5000

int hn_put(object,param,value)
hn_any *object;
char *param;
hn_any *value;
{
	hn_any *any;

	if (!hn_ok()) 
		return(0);
	if (!hn_valid_object(object)) {
		hn_error("hn_put: invalid object");
		return(0);
	}

	hn_write(object);
	hn_flush();
	hnps_name(param);
	hn_write(value);
	hnps_put();
	return(1);
}

hn_any *hn_get(object,param)
hn_any *object;
char *param;
{
	hn_any *any;

	if (!hn_ok()) 
		return(NULL);
	if (!hn_valid_object(object)) {
		hn_error("hn_get: invalid object");
		return(NULL);
	}

	hn_write(object);
	hn_flush();
	hnps_name(param);
	hnps_get();
	hnps_PrintClient();
	return(hn_read_sync(RESULTDELAY));
}

int hn_exists(object)
hn_any *object;
{
	hn_any *any;

	if (!hn_ok()) 
		return(0);
	if (!hn_valid_object(object))
		return(0);
	hn_write(object);
	hn_flush();
	hnps_checknull();
	hnps_PrintClient();
	return(hn_read_sync(RESULTDELAY) == hn_false);
}

int hn_send(target,message,args)
hn_any *target;
char *message;
hn_any *args;
{
	if (!hn_ok()) 
		return(0);
	if (!hn_valid_object(target)) {
		hn_error("hn_send: invalid target");
		return(0);
	}

	if (args) {
		if (args->type != hn_array_type) {
			hnps_begin_array();
			hn_write(args);
			hnps_end_array();
		} else
			hn_write(args);
	} else
		hn_write(hn_nullarray);

	hnps_name(message);
	hn_write(target);
	hnps_Send(1);
	ps_flush_PostScript();
	return(1);
}

int hn_show(object)
hn_any *object;
{
	if (!hn_ok())
		return(0);
	if (hn_valid_object(object)) {
		if (object->u.object.type == hn_Stack) {
			hn_write(object);
			hnps_ShowStack();
			return(1);
		} else 
			return(hn_send(object,"Show",NULL));
	} else
		return(0);
}

int hn_hide(object)
hn_any *object;
{
	if (!hn_ok())
		return(0);
	if (hn_valid_object(object)) {
		if (object->u.object.type == hn_Stack) {
			hn_write(object);
			hnps_HideStack();
			return(1);
		} else 
			return(hn_send(object,"Hide",NULL));
	} else
		return(0);
}

int hn_connect(object)
hn_any *object;
{
	if (hn_ok() && hn_valid_object(object))
		if (object->u.object.type == hn_Stack)
			return(hn_send(object,"Connect",NULL));
	return(0);
}

int hn_set_canvas(object)
hn_any *object;
{
	if (hn_ok() && hn_valid_object(object)) {
		if (object->u.object.type == hn_Canvas) {
			hn_write(object);
			hnps_set_canvas();
			return(1);
		}
	}
	return(0);
}

int hn_set_drawing(object,file)
hn_any *object;
char *file;
{
	if (hn_ok() && hn_valid_object(object)) {
		if (object->u.object.type == hn_Button) {
			hnps_string(file);
			hn_write(object);
			hnps_set_draw();
			return(1);
		}
	}
	return(0);
}

hn_any *hn_rename(object,name)
hn_any *object;
char *name;
{
	if (hn_ok() && hn_valid_object(object)) {
		if (object->u.object.type != hn_Stack) {
			hnps_name(name);
			hn_write(object);
			hnps_rename();
			hnps_PrintClient();
			return(hn_read_sync(RESULTDELAY));
		}
	} 
	return(hn_null);
}

hn_any *hn_stack_copy(object)
hn_any *object;
{
	if (hn_ok() && hn_valid_object(object)) {
		if (object->u.object.type == hn_Stack) {
			hnps_name(object->u.object.name);
			hnps_stack_copy();
			hnps_PrintClient();
			return(hn_read_sync(RESULTDELAY));
		}
	}
	return(hn_null);
}

int hn_show_error(text)
hn_any *text;
{
	if (hn_ok()) {
		if (text->type == hn_array_type) {
			hn_write(text);
			hnps_show_error();
			return(1);
		}
	}
	return(0);
}

int hn_ps(postscript)
char *postscript;
{
	if (hn_ok()) {
		fprintf(PostScript,postscript);
		return(1);
	} else
		return(0);
}
