
#ifndef lint
static char sccsid[] = "@(#)io.c 3.3 9/21/90 Arthur van Hoff (Turing Institute)";
#endif lint

/****************************************************************************
file: 	io.c		(c)1989 Turing Institute
author:	Arthur van Hoff

Description:
	Provide I/O utillities for HyperNeWS client programs.

	void hn_flush()
		Flush pending output to HyperNeWS.

	void hn_flush_input()
		Flush the HyperNeWS event queue, hn_ioq.

	hn_any *hn_read(int delay)
		Read data from HyperNeWS, hn_timeout is returned when
		the delay expires. hn_ioerror is returned when the
		connection with HyperNeWS is lost. The delay is given
		in micro seconds.
	
	hn_any *hn_read_sync(int delay)
		Will return the data which follows the first hn_sync message.
		This routine should be used after the PostScript routine
		PrintClient is called.
	
	int hn_write(hn_any *any)
		Write data to HyperNeWS. 0 is returned when the connection
		with HyperNeWS is lost, 1 otherwise.
*/

#include <stdio.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/types.h>

#include "hnps.h"
#include "token.h"
#include "hypernews.h"

#ifndef PSFILE
#define         filenr(fp)      fileno(fp)
#else
#define         filenr(fp)      psio_fileno(fp)
#endif PSFILE

#ifndef BUFSIZ
#define	BUFSIZ	512
#endif BUFSIZ

#define PENDING	-3
#define TIMEOUT	-2
#define STRLEN	256

#define hn_get()	((incount-- > 0) ? (*inptr++) : hn_fill_buffer())

struct hn_ioq {
	hn_any *data;
	struct hn_ioq *next;
} *hn_ioq = NULL;

extern double atof();

static char		strbuf[STRLEN];
static unsigned char	inbuf[BUFSIZ];
static unsigned char	*inptr = inbuf;
static int		incount = 0;
static int		hn_delay = hn_forever;
int			hn_status = 1;
int			hn_mask = 0;

/****************************************************************************/

/* A version of select(2) that keeps going on interrupts from signals.
 */
static int select_signal(width, readfds, writefds, exceptfds, timeout)
int width;
fd_set *readfds, *writefds, *exceptfds;
struct timeval *timeout;
{
	int n;

	while (1) {
		errno = 0;
		n = select(width, readfds, writefds, exceptfds, timeout); 
		if ((n >= 0) || (errno != EINTR))
			return n;
	}
}

/****************************************************************************/

static int hn_fill_buffer()
{
	int	mask0, mask, n;
	struct timeval tm;


	if (!hn_ok())
		return(EOF);

	tm.tv_sec = hn_delay / 1000;
	tm.tv_usec = hn_delay % 1000 * 1000;

	mask0 = (1<<filenr(PostScriptInput));
	mask = mask0 | hn_mask;
	if ((n = select_signal(32,&mask,0,0,(hn_delay == hn_forever) ? 0 : &tm)) < 0) {
		if (hn_verbose)
			fprintf(stderr,"hn_read: select failed\n");
		hn_status = 0;
		return(EOF);
	}
	if ((n == 0) && (hn_delay != hn_forever))
		return(TIMEOUT);
	if (!(mask & mask0)) {
		/* input on user file */
		return(PENDING);
	}
	if ((incount = read(filenr(PostScriptInput),inbuf,BUFSIZ)) <= 0) {
		if (hn_verbose)
			fprintf(stderr,"hn_read: read error\n");
		hn_status = 0;
		return(EOF);
	}
	inptr = inbuf;
	incount--;
	return(*inptr++);
}

/****************************************************************************/

void hn_flush()
{
	ps_flush_PostScript();
}

void hn_flush_input()
{
	struct hn_any *any;

	ps_flush_PostScript();

	while (any = hn_read(0)) {
		if (any == hn_timeout)
			return;
		if (any == hn_ioerror)
			return;
		if (any == hn_quit)
			return;
		if (any == hn_pending)
			return;
		hn_free(any);
	}
}

/****************************************************************************/

hn_any *hn_read0(delay)
int delay;
{
	register int	n;
	register short	c;
	register char	*p, *name;
	register hn_any *any, *target, *args;

	if (!hn_ok())
		return(hn_ioerror);

	ps_flush_PostScript();

	hn_delay = delay;

	while ((c = hn_get()) >= 0) {
		if (c & 0200) {
			switch (c &= 0177) {
			case 0:
				return(hn_null);
			case 1:
				return(hn_true);
			case 2:
				return(hn_false);
			case 3:
				c = hn_get();
				p = strbuf;
				while (c--)
					*p++ = hn_get();
				*p = '\0';
				return(hn_new_name(strbuf));
			case 4:
				c = hn_get();
				p = strbuf;
				while (c--)
					*p++ = hn_get();
				*p = '\0';
				return(hn_new_string(strbuf));
			case 5:
				c = hn_get() << 8;
				c |= hn_get();
				return(hn_new_number((double)c));
			case 6:
				c = hn_get();
				p = strbuf;
				while (c--)
					*p++ = hn_get();
				*p = '\0';
				return(hn_new_number(atof(strbuf)));
			case 7:
				c = hn_get();
				p = strbuf;
				while (c--)
					*p++ = hn_get();
				*p = '\0';
				return(hn_new_object(NULL,hn_Stack,strbuf));
			case 8:
				c = hn_get();
				p = strbuf;
				while (c--)
					*p++ = hn_get();
				*p++ = '\0';
				n = hn_get();
				name = p;
				c = hn_get();
				while (c--)
					*p++ = hn_get();
				*p++ = '\0';
				return(hn_new_object(strbuf,n,name));
			case 9:
				c = hn_get() << 8;
				c |= hn_get();
				any = hn_new_array_body((int)c);
				for (n = 0 ; n < c ; n++)
					any->u.array.argv[n] = hn_read0(delay);
				return(any);
			case 10:
				c = hn_get();
				p = strbuf;
				while (c--)
					*p++ = hn_get();
				*p++ = '\0';
				name = str2name(strbuf);
				args = hn_read0(delay);
				target = hn_read0(delay);
				return(hn_new_message(target,name,args));
			case 11:
				hn_status = 0;
				return(hn_quit);
			case 12:
				return(hn_sync);
			default:
				if (hn_verbose)
					fprintf(stderr,"funny escape: %d\n",c);
				break;
			}
		} else {
			if (hn_verbose)
				fprintf(stderr,"funny character: '%c' (%d)\n",c,c);
		}
	}
	switch (c) {
	case PENDING:
		return(hn_pending);
	case TIMEOUT:
		return(hn_timeout);
	case EOF:
	default:
		return(hn_ioerror);
	}
}

hn_any *hn_read(delay)
int delay;
{
	struct hn_any *any;
	struct hn_ioq *p;

	if (hn_ioq) {
		any = hn_ioq->data;
		p = hn_ioq;
		hn_ioq = hn_ioq->next;
		hn_destroy(p);
		return(any);
	} else
		return(hn_read0(delay));
}

hn_any *hn_read_sync(delay)
int delay;
{
	register struct hn_any *any;
	register struct hn_ioq **p = &hn_ioq;

	ps_flush_PostScript();

	while (*p)
		p = &(*p)->next;

	while (hn_ok()) {
		any = hn_read0(delay);
		if (any == hn_timeout)
			return(any);
		if (any == hn_ioerror)
			return(any);
		if (any == hn_quit)
			return(any);
		if (any == hn_sync) {
			any = hn_read0(delay);
			if (hn_verbose) {
				fprintf(stderr,"sync read:");
				hn_print(stderr,any);
				fprintf(stderr,"\n");
			}
			return(any);
		}
		/* q this */
		if (hn_verbose) {
			fprintf(stderr,"queued: ");
			hn_print(stderr,any);
			fprintf(stderr,"\n");
		}
		*p = (struct hn_ioq *)hn_alloc(sizeof(struct hn_ioq));
		(*p)->data = any;
		(*p)->next = NULL;
		p = &(*p)->next;
	}
	return(hn_ioerror);
}

/****************************************************************************/

int hn_write(any)
hn_any *any;
{
	register int n;

	if (!hn_ok())
		return(0);

	if (!any) {
		hnps_null();
		return(1);
	}
	switch (any->type) {
	case hn_null_type:
		hnps_null();
		return(1);
	case hn_boolean_type:
		if (any->u.boolean)
			hnps_true();
		else
			hnps_false();
		return(1);
	case hn_number_type:
		if (any->u.number == (int)(any->u.number))
			hnps_integer((int)any->u.number);
		else
			hnps_float(any->u.number);
		return(1);
	case hn_name_type:
		hnps_name(any->u.name);
		return(1);
	case hn_string_type:
		hnps_string(any->u.string);
		return(1);
	case hn_array_type:
		if (any->u.array.argc) {
			hnps_begin_array();
			for (n = 0 ; n < any->u.array.argc ; n++)
				hn_write(any->u.array.argv[n]);
			hnps_end_array();
		} else
			hnps_nullarray();
		return(1);
	case hn_message_type:
		hn_write(any->u.message.args);
		hnps_name(any->u.message.name);
		hn_write(any->u.message.target);
		hnps_Send();
		ps_flush_PostScript();
		return(1);
	case hn_object_type:
		if (any->u.object.type == hn_Stack) {
			hnps_name(any->u.object.name);
			hnps_FindStack();
		} else {
			hnps_name(any->u.object.stack);
			hnps_name(any->u.object.name);
			hnps_FindObject();
		}
		return(1);
	default:
		hn_error("hn_write: object type error");
	}
}
