
#ifndef lint
static char sccsid[] = "@(#)main.c 3.1 3/19/90 Arthur van Hoff (Turing Institute)";
#endif lint

/*****************************************************************
file: main.c		(c)1989 Turing Institute
authour: Arthur van Hoff

Description:
	Main interface routines for HyperNeWS clients.

	hn_error(char *str)
		Print an error on stderr.

	hn_register(int (*)() handler, hn_any *match)
		Register a procedure. The procedure will be called by
		hn_listen (see below) when the input from HyperNeWS matches
		the 'match' argument. Any part of the 'match' data structure
		can be set to ANY which will match with anything.
		If the handler returns 0 other handlers will be considered.
		If it returns 1 NO other handlers will be considered for
		the input.

	hn_listen(int delay)
		Listen for input. When input is received a list of registered
		handlers is checked and the appropriate handler is called.
		The input is passed as the first argument to the handler.
		When the input is of the hn_message_type type, a second argument
		(the argument count) and a third argument (the argument vector)
		are passed on as well.

	int hn_start(char *programname)
		Connect the client to HyperNeWS using its processid and
		the programname. This routine should be called before you
		can communicate to HyperNeWS or use any of the data-structures.
		Returns 1 if the connection was established, 0 otherwise.

	int hn_stop()
		Disconnect the client from HyperNeWS. Always returns 1.

*/

#include <stdio.h>
#include "hnps.h"
#include "token.h"
#include "hypernews.h"

extern struct hn_ioq {
	hn_any *data;
	struct hn_ioq *next;
} *hn_ioq;

extern int getpid();

typedef struct hn_listen_entry {
        struct hn_listen_entry  *next;
        hn_any                  *match;
        int                     (*handler)();
} hn_listen_entry;

static hn_listen_entry *hn_listen_q = NULL;

int 		hn_running = 0;
int		hn_verbose = 0;
int		hn_intro = 0;
static int 	hn_initialized = 0;

char 		*hn_object_type_names[hn_nr_object_types+1];

hn_any  *hn_true = NULL,
        *hn_false = NULL,
        *hn_null = NULL,
        *hn_ioerror = NULL,
        *hn_timeout = NULL,
        *hn_quit = NULL,
        *hn_nullarray = NULL,
        *hn_sync = NULL,
        *hn_pending = NULL;


hn_error(str)
char *str;
{
	fprintf(stderr,"hn_error: %s\n",str);
}

void hn_register(handler,match)
int (*handler)();
hn_any *match;
{
        register hn_listen_entry **p = &hn_listen_q;

        while (*p)
                p = &(*p)->next;
        
        *p = (hn_listen_entry *)hn_alloc(sizeof(hn_listen_entry));
        (*p)->next = NULL;
        (*p)->match = match;
        (*p)->handler = handler;
}

int hn_listen(delay)
int delay;
{
        register hn_listen_entry *p;
        register hn_any *any;
        register int n;

        if (!hn_listen_q)
                return(0);
        while (1) {
                any = hn_read(delay);
		if (hn_verbose) {
			fprintf(stderr,"hn_listen: ");
			hn_print(stderr,any);
			fprintf(stderr,"\n");
		}
                n = 0;
                p = hn_listen_q;
                while (p) {
                        if (hn_match(p->match,any)) {
                                n++;
				if (any->type == hn_message_type) {
					if ((*p->handler)(any,
						any->u.message.args->u.array.argc,
						any->u.message.args->u.array.argv))
						break;
				} else {
					if ((*p->handler)(any))
						break;
				}
                        }
                        p = p->next;
                }
                if ((!n) && (any == hn_ioerror))
                        return(EOF);
                if (any == hn_quit)
                        return(1);
                hn_free(any);
        }
}
 
int hn_start(progname)
char *progname;
{
	if (!hn_initialized) {
		/* setup some constants */
		hn_constant(hn_null = (hn_any *)hn_alloc(sizeof(hn_any)));
		hn_null->type = hn_null_type;

		hn_constant(hn_true = (hn_any *)hn_alloc(sizeof(hn_any)));
		hn_true->type = hn_boolean_type;
		hn_true->u.boolean = 1;

		hn_constant(hn_false = (hn_any *)hn_alloc(sizeof(hn_any)));
		hn_false->type = hn_boolean_type;
		hn_false->u.boolean = 0;


		hn_constant(hn_nullarray = (hn_any *)hn_alloc(sizeof(hn_any)));
		hn_nullarray->type = hn_array_type;
		hn_nullarray->u.array.argc = 0;
		hn_nullarray->u.array.argv = (hn_any **)hn_alloc(0);

		hn_constant(hn_quit = hn_new_name("quit"));
		hn_constant(hn_timeout = hn_new_name("timeout"));
		hn_constant(hn_ioerror = hn_new_name("ioerror"));
		hn_constant(hn_sync = hn_new_name("sync"));
		hn_constant(hn_pending = hn_new_name("pending"));

		hn_object_type_names[0] =
			str2name("unknown");
		hn_object_type_names[(int)hn_Stack] =
			str2name("Stack");
		hn_object_type_names[(int)hn_BackGround] =
			str2name("BackGround");
		hn_object_type_names[(int)hn_Card] =
			str2name("Card");
		hn_object_type_names[(int)hn_Button] =
			str2name("Button");
		hn_object_type_names[(int)hn_EditText] =
			str2name("EditText");
		hn_object_type_names[(int)hn_Slider] =
			str2name("Slider");
		hn_object_type_names[(int)hn_PullDown] =
			str2name("PullDown");
		hn_object_type_names[(int)hn_ColorSelect] =
			str2name("ColorSelect");
		hn_object_type_names[(int)hn_Canvas] =
			str2name("Canvas");

		hn_initialized = 1;
	}
	if (!hn_running) {
		if (hn_verbose)
			fprintf(stderr,"hn_start: starting\n");
		if (!ps_open_PostScript()) {
			hn_error("hn_start: cannot connect to news_server");
			return(0);
		}
		hn_running = 1;
		hnps_start((progname ? progname : "unknown"),
				getpid(),hn_intro);
		ps_flush_PostScript();
		if (hn_read(20000) != hn_sync) {
			hn_running = 0;
			hn_error("hn_start: timeout");
			return(0);
		}
		if (hn_verbose)
			fprintf(stderr,"hn_start\n");
	}
	return(1);
}

int hn_stop()
{
	if (hn_running) {
		hnps_stop();
		ps_close_PostScript();
		hn_running = 0;
		if (hn_verbose)
			fprintf(stderr,"hn_stop\n");
	}
	hn_ioq = 0;
	hn_listen_q = 0;
	return(1);
}
