
#ifndef lint
static char sccsid[] = "@(#)util.c 3.1 3/19/90 Arthur van Hoff (Turing Institute)";
#endif lint

/****************************************************************************
file: 	util.c		(c)1989 Turing Institute
author:	Arthur van Hoff

Description:
	Utillity routines for HyperNeWS client programs.

	char *str2name(char *str)
		Provide a hashed index for strings.

	hn_any *hn_new_number(double n)
	hn_any *hn_new_string(char *str)
	hn_any *hn_new_name(char *str)
	hn_any *hn_new_object(char *stack, int type, char *name)
	hn_any *hn_new_array_body(int length)
	hn_any *hn_new_array(hn_any *obj1, *obj2, *obj3, ....., NULL)
	hn_any *hn_new_message(hn_any *target, char *name, hn_any *args)
		Some routines to create new objects.

	int hn_match(hn_any *a, hn_any *b)
		Returns 1 if a and b match, 0 otherwise. Any NULL branch
		of the data structures matches anything.

	void hn_free(hn_any *p)
		Delete an entire data structure.

	void hn_print(FILE *fp, hn_any *p)
		Print a datastructure to a file.
*/

#include <stdio.h>
#include <varargs.h>
#include "hnps.h"
#include "hypernews.h"
#include "qlalloc.h"

#define HASHTABSIZ	513

extern void *malloc();
extern void free();
extern int strlen();
extern int strcmp();
extern void strcpy();

typedef struct name {
	struct name *next;
	char	data[2];
} name;

static name **hashtab = NULL;

/****************************************************************************/

void *hn_alloc(bytes)
int bytes;
{
	return(qlalloc(bytes));
}

void hn_destroy(ptr)
void *ptr;
{
	qlfree(ptr);
}

/****************************************************************************/

char *str2name(str)
register char *str;
{
	register name *p;
	register int len = strlen(str);

	if (!hashtab)
		hashtab = (name **)hn_alloc(sizeof(name *) * HASHTABSIZ);

	p = (name *)&hashtab[(str[0] + (str[len-1] * len)) % HASHTABSIZ];
	while (p->next) {
		p = p->next;
		if (!strcmp(p->data,str))
			return(p->data);
	}

	p->next = (name *)hn_alloc(sizeof(name *) + len + 1);
	p = p->next;
	p->next = NULL;
	strcpy(p->data,str);
	return(p->data);
}

/****************************************************************************/

hn_any *hn_new_number(value)
double value;
{
	register hn_any *any = (hn_any *)hn_alloc(sizeof(hn_any));
	any->refcount = 0;
	any->type = hn_number_type;
	any->u.number = value;
	return(any);
}

hn_any *hn_new_string(value)
char *value;
{
	register hn_any *any = (hn_any *)hn_alloc(sizeof(hn_any));
	any->refcount = 0;
	any->type = hn_string_type;
	any->u.string = str2name(value);
	return(any);
}

hn_any *hn_new_name(value)
char *value;
{
	register hn_any *any = (hn_any *)hn_alloc(sizeof(hn_any));
	any->refcount = 0;
	any->type = hn_name_type;
	any->u.name = str2name(value);
	return(any);
}

hn_any *hn_new_object(stack,type,name)
char *stack;
hn_type type;
char *name;
{
	register hn_any *any = (hn_any *)hn_alloc(sizeof(hn_any));
	any->refcount = 0;
	any->type = hn_object_type;
	any->u.object.stack = (stack ? str2name(stack) : NULL);
	any->u.object.type = type;
	any->u.object.name = str2name(name);
	return(any);
}

hn_any *hn_new_array_body(argc)
int argc;
{
	register hn_any *any;

	if (!argc)
		return(hn_nullarray);
	any = (hn_any *)hn_alloc(sizeof(hn_any));
	any->refcount = 0;
	any->type = hn_array_type;
	any->u.array.argc = argc;
	any->u.array.argv = (hn_any **)hn_alloc(sizeof(hn_any *) * argc);
	return(any);
}

hn_any *hn_new_array(va_alist)
va_dcl
{
	register hn_any *any;
	register int argc = 0;
	register va_list alist;
	register hn_any *p;

	va_start(alist);
	while (va_arg(alist,hn_any *))
		argc++;
	va_end(alist);

	if (!argc)
		return(hn_nullarray);
	any = (hn_any *)hn_alloc(sizeof(hn_any));
	any->refcount = 0;
	any->type = hn_array_type;
	any->u.array.argc = argc;
	any->u.array.argv = (hn_any **)hn_alloc(sizeof(hn_any *) * argc);

	va_start(alist);
	argc = 0;
	while (p = va_arg(alist,hn_any *))
		any->u.array.argv[argc++] = p;
	va_end(alist);

	return(any);
}

hn_any *hn_new_message(target,name,args)
hn_any *target;
char *name;
hn_any *args;
{
	register hn_any *any = (hn_any *)hn_alloc(sizeof(hn_any));

	any->refcount = 0;
	any->type = hn_message_type;
	any->u.message.target = target;
	any->u.message.name = (name ? str2name(name) : NULL);
	any->u.message.args = args;
	return(any);
}

/****************************************************************************/

int hn_match(any1,any2)
register hn_any *any1, *any2;
{
	register int n;

	if (any1 == any2)
		return(1);
	if ((!any1) || (!any2))
		return(1);
	if (any1->type != any2->type)
		return(0);
	switch (any1->type) {
	case hn_boolean_type:
		return(any1->u.boolean == any2->u.boolean);
	case hn_null_type:
		return(1);
	case hn_number_type:
		return(any1->u.number == any2->u.number);
	case hn_string_type:
		return(!strcmp(any1->u.string,any2->u.string));
	case hn_name_type:
		return(any1->u.name == any2->u.name);
	case hn_object_type:
		if (any1->u.object.type != any2->u.object.type) {
			if (any1->u.object.type && any2->u.object.type)
				return(0);
		}
		if (any1->u.object.stack != any2->u.object.stack) {
			if (any1->u.object.stack && any2->u.object.stack)
				return(0);
		}
		if (any1->u.object.name != any2->u.object.name) {
			if (any1->u.object.name && any2->u.object.name)
				return(0);
		}
		return(1);
	case hn_array_type:
		if (any1->u.array.argc != any2->u.array.argc)
			return(0);
		for (n = 0 ; n < any1->u.array.argc ; n++)
			if (!hn_match(any1->u.array.argv[n],any2->u.array.argv[n]))
				return(0);
		return(1);
	case hn_message_type:
		if (!hn_match(any1->u.message.target,any2->u.message.target))
			return(0);
		if (!hn_match(any1->u.message.args,any2->u.message.args))
			return(0);
		if (any1->u.message.name != any2->u.message.name) {
			if (any1->u.message.name && any2->u.message.name)
				return(0);
		}
		return(1);
	default:
		hn_error("hn_match: object type error");
		return(0);
	}
}

/****************************************************************************/

void hn_free(any)
register hn_any *any;
{
	register int n;

	if (!any)
		return;
	if (any->refcount >= 128)
		return;
	if (any->refcount-- > 0)
		return;

	switch (any->type) {
	case hn_boolean_type:
	case hn_null_type:
		/* should never happen */
	case hn_number_type:
	case hn_string_type:
	case hn_name_type:
	case hn_object_type:
		break;
	case hn_array_type:
		for (n = 0 ; n < any->u.array.argc ; n++)
			hn_free(any->u.array.argv[n]);
		hn_destroy(any->u.array.argv);
		break;
	case hn_message_type:
		hn_free(any->u.message.target);
		hn_free(any->u.message.args);
		break;
	default:
		hn_error("hn_free: object type error");
		return;
	}
	hn_destroy(any);
}

/****************************************************************************/

void hn_print(fp,any)
register FILE *fp;
register hn_any *any;
{
	register int n;
	register char *p;

	if (!any) {
		fprintf(fp,"NIL");
		return;
	}
	switch (any->type) {
	case hn_number_type:
		fprintf(fp,"%g",any->u.number);
		break;
	case hn_boolean_type:
		fprintf(fp,(any->u.boolean ? "true" : "false"));
		break;
	case hn_string_type:
		fprintf(fp,"(");
		p = any->u.string;
		while (*p) {
			switch (*p) {
			case '\n':
			case '\t':
			case '(':
			case ')':
				fprintf(fp,"\\");
			default:
				fprintf(fp,"%c",*p++);
			}
		}
		fprintf(fp,")");
		break;
	case hn_name_type:
		fprintf(fp,"/%s",any->u.name);
		break;
	case hn_null_type:
		fprintf(fp,"null");
		break;
	case hn_object_type:
		if (any->u.object.stack)
			fprintf(fp,"%s:",any->u.object.stack);
		fprintf(fp,"%s(%s)",
			hn_object_type_names[(int)any->u.object.type],
			any->u.object.name);
		break;
	case hn_array_type:
		fprintf(fp,"[");
		for (n = 0 ; n < any->u.array.argc ; n++) {
			hn_print(fp,any->u.array.argv[n]);
			if ((n+1) < any->u.array.argc)
				fprintf(fp," ");
		}
		fprintf(fp,"]");
		break;
	case hn_message_type:
		fprintf(fp,"message(");
		hn_print(fp,any->u.message.target);
		if (any->u.message.name)
			fprintf(fp,",%s,",any->u.message.name);
		else
			fprintf(fp,",NIL,");
		hn_print(fp,any->u.message.args);
		fprintf(fp,")");
		break;
	default:
		hn_error("hn_print: object type error");
		return;
	}
}
