
% "@(#)demo.pl 3.2 9/21/90 Arthur van Hoff (Turing Institute)";

%######################################################################
% file: 	demo.pl			(c)1989 Turing Institute
% author: 	Arthur van Hoff
%
% Description:
% 	Use the HyperNeWS interface to implement an interactive
%	prolog help system.
%

:-
	ensure_loaded([
		'interface.pl',
		library(basics),
		library(lineio),
		library(charsio)
	]).

%######################################################################
% help stuff

:-
	dynamic current_help/1, file_read/2.

help_directory('/usr/local/quintus/generic/q2.5/helpsys').

aformat(Atom,Format,Args) :-
	with_output_to_chars(format(Format,Args),Chars),
	name(Atom,Chars).

lookfor(Atom,Text) :-
	file_read(Atom,Text), !.
lookfor(Atom,['','     Prolog cross-reference    '|Ref]) :-
	help_directory(Dir),
	aformat(A,'egrep -i ''~a'' ~a/menuindex > ~~/tmp.grep',[Atom,Dir]),
	unix(shell(A)),
	read_file('~/tmp.grep',Ref),
	retract(file_read('~/tmp.grep',_)),
	asserta(file_read(Atom,Ref)),
	unix(shell('rm ~/tmp.grep')).

%######################################################################
% reading files

read_help_file(File,Text) :-
	help_directory(Dir),
	aformat(Full,'~a/~a',[Dir,File]),
	read_file(Full,Text).

read_file(File,Text) :-	
	file_read(File,Text), !.
read_file(File,Text) :-	
	hn_set_text('PrologHelp':edittext(title),'reading...'),
	nofileerrors,
	open(File,read,Stream),
	current_input(Old),
	set_input(Stream),
	read_lines(0,Text),
	set_input(Old),
	close(Stream), 
	asserta(file_read(File,Text)),
	fileerrors, !.
read_file(_,_) :-
	fileerrors, !, fail.

read_lines(1000,['**** More lines (ignored) ****']).
read_lines(N,[Ln|Rest]) :-
	get_line(Chars,10),
	format_line(Chars,CharsFormat),
	name(Ln,CharsFormat),
	N1 is N + 1,
	read_lines(N1,Rest).
read_lines(_,[]).

format_line(L1,L2) :-
	append(Head,[123,109,97,110,117,97,108|Rest],L1),
	append(Between,[125|End],Rest), !,
	format_line(End,EndFormat),
	append(Head,[1,2|"manual"],H1),
	append(H1,Between,H2),
	append(H2,[1,1|EndFormat],L2).
format_line(X,X).

section2file(Atom,Atom) :-
	atom(Atom).
section2file(Atom,File) :-
	section2file0(Atom,File).

section2file0(user,'U').
section2file0(ref,'R').
section2file0(sysdep,'S').
section2file0(lib,'L').
section2file0(H-N,R2) :-
	integer(N), N < 10,
	section2file(H,R1),
	aformat(R2,'~a0~d',[R1,N]).
section2file0(H-N,R2) :-
	integer(N), N >= 10,
	section2file(H,R1),
	aformat(R2,'~a~d',[R1,N]).

%######################################################################
% Moving around

next_section(manuals,user).
next_section(user,ref).
next_section(ref,sysdep).
next_section(sysdep,lib).
next_section(lib,manuals).

next_section(Sec-N1-N2-N3,Sec-N1-N2-N4) :-
	N4 is N3+1, section_ok(Sec-N1-N2-N4), !.
next_section(Sec-N1-N2-_,Sec-N1-N2) :- 
	!.
next_section(Sec-N1-N2,Sec-N1-N3) :-
	N3 is N2 + 1, section_ok(Sec-N1-N3), !.
next_section(Sec-N1-_,Sec-N1) :-
	!.
next_section(Sec-N1,Sec-N2) :-
	N2 is N1 + 1, section_ok(Sec-N2), !.
next_section(Sec-_,Sec) :-
	!.
next_section(Sec,Sec).

section_ok(Sec) :-
	section2file(Sec,File), 
	nofileerrors,
	help_directory(Dir),
	aformat(Full,'~a/~a',[Dir,File]),
	open(Full,read,S),
	close(S), !.

prev_section(manuals,lib).
prev_section(user,manuals).
prev_section(ref,user).
prev_section(sysdep,ref).
prev_section(lib,sysdep).

prev_section(Sec-N1-N2-N3,Sec-N1-N2-N4) :-
	N4 is N3-1, section_ok(Sec-N1-N2-N4), !.
prev_section(Sec-N1-N2-_,Sec-N1-N2) :- 
	!.
prev_section(Sec-N1-N2,Sec-N1-N3) :-
	N3 is N2-1, section_ok(Sec-N1-N3), !.
prev_section(Sec-N1-_,Sec-N1) :-
	!.
prev_section(Sec-N1,Sec-N2) :-
	N2 is N1-1, section_ok(Sec-N2), !.
prev_section(Sec-_,Sec) :-
	!.
prev_section(Sec,Sec).

%######################################################################
% HyperNeWS interface.

start_help :-
	hn_start, 
	hn_connect(stack('PrologHelp')),
	show_text(manuals),
	hn_show(stack('PrologHelp')),
	hn_run.

current_help(none).

show_text(Section) :-
	current_help(Section), !.
show_text(Section) :-
	with_output_to_chars(write(Section),Chars),
	name(Sec,Chars),
	section2file(Section,File),
	read_help_file(File,[_,Title,_|T]),
	strip_spaces(Title,Tit),
	append(Text,[_,_],T),
	retractall(current_help(_)),
	assert(current_help(Section)),
	length(Text,Len),
	aformat(Lines,'~d lines',[Len]),
	hn_set_text('PrologHelp':edittext(title),Tit),
	hn_set_text('PrologHelp':edittext(section),Sec),
	hn_set_text('PrologHelp':edittext(lines),Lines),
	hn_set_text('PrologHelp':edittext(contents),text(Text)).
show_text(X) :-
	write('FAIL: '), write(X), nl.

strip_spaces(T1,T2) :-
	name(T1,S1),
	strip_space(S1,S2),
	name(T2,S2).

strip_space([32|L1],L2) :-
	strip_space(L1,L2).
strip_space(L,L).

hn_message(ioerror) :-
	halt.
hn_message(quit) :-
	halt.
hn_message('PrologHelp':button('NextButton'),'Action',_) :-
	current_help(Sec), next_section(Sec,Next), show_text(Next).
hn_message('PrologHelp':button('PreviousButton'),'Action',_) :-
	current_help(Sec), prev_section(Sec,Prev), show_text(Prev).
hn_message('PrologHelp':_,'WordAction',[string(Word)]) :-
	name(Word,S1),
	% "\1\2manual("
	append(_,[1,2,109,97,110,117,97,108,40|S2],S1),
	% ")\1\1"
	append(S3,[41,1,1|_],S2),
	append(S3,".",S4),
	with_input_from_chars(read(Manual),S4),
	show_text(Manual).
hn_message(_,'Find',[string(Word)]) :-
	hn_hide(stack('FindString')),
	aformat(Tmp,'Searching for ''~a''...',[Word]),
	hn_set_text('PrologHelp':edittext(title),Tmp),
	lookfor(Word,[_,_|Text]),
	retractall(current_help(_)),
	assert(current_help(none)),
	length(Text,Len),
	aformat(Lines,'~d lines',[Len]),
	hn_set_text('PrologHelp':edittext(title),'Prolog cross-reference'),
	hn_set_text('PrologHelp':edittext(section),Word),
	hn_set_text('PrologHelp':edittext(lines),Lines),
	hn_set_text('PrologHelp':edittext(contents),text(Text)).
hn_message(_,'Find',[string(Word)]) :-
	aformat(Tmp,'''~a'' not found',[Word]),
	hn_set_text('PrologHelp':edittext(title),Tmp), !.
hn_message(_,'GoHome',_) :-
	show_text(manuals).
