
% "@(#)hnintf.pl 3.1 3/19/90 Arthur van Hoff (Turing Institute)";

%######################################################################
% file:         interface.ps            (c)1989 Turing Institute
% author:       Arthur van Hoff
%
% Description:
%       Define an low level interface between Quintus Prolog and HyperNeWS.
%	See 'interface.pl' for a higher level interface.
%

:-
	ensure_loaded([
		hnlibdir
	]).

%######################################################################
% foreign stuff (ignore everything that starts with 'hnpr_').

foreign_file('interface.o',Funcs) :-
	setof(X,Y^foreign(X,Y),Funcs).

foreign(hnpr_start,
		hnpr_start(-string,+string)).
foreign(hnpr_stop,
		hnpr_stop(-string)).
foreign(hnpr_ok,
		hnpr_ok(-string)).
foreign(hnpr_flush,
		hnpr_flush(-string)).
foreign(hnpr_flush_input,
		hnpr_flush_input(-string)).
foreign(hnpr_verbose,
		hnpr_verbose(+integer)).
foreign(hnpr_clear,
		hnpr_clear(-string)).
foreign(hnpr_ps,
		hnpr_ps(-string,+string)).

foreign(hnpr_write,
		hnpr_write(-string)).
foreign(hnpr_read,
		hnpr_read(-string,+integer)).

foreign(hnpr_put_any,
		hnpr_put_any(-string,+string)).
foreign(hnpr_get_any,
		hnpr_get_any(-string,+string)).
foreign(hnpr_show,
		hnpr_show(-string)).
foreign(hnpr_hide,
		hnpr_hide(-string)).
foreign(hnpr_exists,
		hnpr_exists(-string)).
foreign(hnpr_set_drawing,
		hnpr_set_drawing(-string,+string)).
foreign(hnpr_rename,
		hnpr_rename(-string,+string)).
foreign(hnpr_stack_copy,
		hnpr_stack_copy(-string)).

foreign(hnpr_put_null,
		hnpr_put_null).
foreign(hnpr_put_boolean,
		hnpr_put_type(+integer)).
foreign(hnpr_put_integer,
		hnpr_put_integer(+integer)).
foreign(hnpr_put_float,
		hnpr_put_float(+float)).
foreign(hnpr_put_string,
		hnpr_put_string(+string)).
foreign(hnpr_put_name,
		hnpr_put_name(+string)).
foreign(hnpr_put_message,
		hnpr_put_message(+string)).
foreign(hnpr_put_object,
		hnpr_put_object(+string,+integer,+string)).
foreign(hnpr_put_stack_object,
		hnpr_put_stack_object(+string)).
foreign(hnpr_put_array,
		hnpr_put_array(+integer)).
foreign(hnpr_put_array_element,
		hnpr_put_array_element(+integer)).

foreign(hnpr_get_type,
		hnpr_get_type(-string,-string)).
foreign(hnpr_get_integer,
		hnpr_get_integer(-string,-integer)).
foreign(hnpr_get_float,
		hnpr_get_float(-string,-float)).
foreign(hnpr_get_string,
		hnpr_get_string(-string,-string)).
foreign(hnpr_get_name,
		hnpr_get_name(-string,-string)).
foreign(hnpr_get_message,
		hnpr_get_message(-string,-string)).
foreign(hnpr_get_object,
		hnpr_get_object(-string,-string,-integer,-string)).
foreign(hnpr_get_array_element,
		hnpr_get_array_element(-string,+integer)).

:-
	hnlibdir(Lib),
	load_foreign_files(['interface.o'],[Lib, '-lhn']),
	abolish(foreign_file,2),
	abolish(foreign,2).

%######################################################################
% I/O to and from prolog.

hnpr_put(null) :-
	hnpr_put_null, !.
hnpr_put(true) :-
	hnpr_put_boolean(1), !.
hnpr_put(false) :-
	hnpr_put_boolean(0), !.
hnpr_put(string(X)) :-
	atom(X), !,
	hnpr_put_string(X).
hnpr_put(text([])) :-
	!, hnpr_put_array(0).
hnpr_put(text([X|L])) :-
	!, length([X|L],N),
	hnpr_put_array(N),
	hnpr_put_text([X|L],0).
hnpr_put([]) :-
	!, hnpr_put_array(0).
hnpr_put([X|L]) :-
	!, length([X|L],N),
	hnpr_put_array(N),
	hnpr_put_array_elements([X|L],0).
hnpr_put(X) :-
	atom(X), !,
	hnpr_put_name(X).
hnpr_put(X) :-
	integer(X), !,
	hnpr_put_integer(X).
hnpr_put(X) :-
	float(X), !,
	hnpr_put_float(X).
hnpr_put(stack(Name)) :-
	atom(Name), !,
	hnpr_put_stack_object(Name).
hnpr_put(Stack:Object) :-
	atom(Stack),
	Object =.. [Type,Name],
	atom(Name),
	Type \== stack,
	hnpr_object_type_name(TypeNr,Type), !,
	hnpr_put_object(Stack,TypeNr,Name).
hnpr_put(message(Target,Name,Array)) :-
	hn_islist(Array),
	hn_object(Target), atom(Name), !,
	hnpr_put(Array),
	hnpr_put(Target),
	hnpr_put_message(Name).

hnpr_put_array_elements([],_).
hnpr_put_array_elements([X|L],N) :-
	hnpr_put(X),
	hnpr_put_array_element(N),
	N1 is N + 1,
	hnpr_put_array_elements(L,N1).

hnpr_put_text([],_).
hnpr_put_text([string(X)|L],N) :-
	!, hnpr_put(string(X)),
	hnpr_put_array_element(N),
	N1 is N + 1,
	hnpr_put_text(L,N1).
hnpr_put_text([X|L],N) :-
	atom(X), !,
	hnpr_put(string(X)),
	hnpr_put_array_element(N),
	N1 is N + 1,
	hnpr_put_text(L,N1).

hnpr_get(Result) :-
	hnpr_get_type(ok,Type),
	hnpr_get(Type,X), !,
	Result = X.

hnpr_get(true,true).
hnpr_get(false,false).
hnpr_get(null,null).
hnpr_get(integer,X) :-
	hnpr_get_integer(ok,X).
hnpr_get(float,X) :-
	hnpr_get_float(ok,X).
hnpr_get(string,string(X)) :-
	hnpr_get_string(ok,X).
hnpr_get(name,X) :-
	hnpr_get_name(ok,X).
hnpr_get(message,message(From,Message,Args)) :-
	hnpr_get_message(ok,Message),
	hnpr_get_type(ok,FromType),
	hnpr_get(FromType,From),
	hnpr_get_type(ok,ArgsType),
	hnpr_get(ArgsType,Args).
hnpr_get(object,Object) :-
	hnpr_get_object(ok,Stack,TypeNr,Name),
	hnpr_object_type_name(TypeNr,Type),
	hnpr_object_name(Type,Stack,Name,Object).
hnpr_get(array,Array) :-
	hnpr_get_array(0,Array).

hnpr_get_array(N,[X|L]) :-
	hnpr_get_array_element(ok,N), !,
	hnpr_get_type(ok,Type),
	hnpr_get(Type,X),
	N1 is N + 1,
	hnpr_get_array(N1,L).
hnpr_get_array(_,[]).

hnpr_object_name(stack,_,Name,stack(Name)).
hnpr_object_name(Type,Stack,Name,Stack:Object) :-
	Object =.. [Type,Name].

hnpr_object_type_name(0,any).
hnpr_object_type_name(1,stack).
hnpr_object_type_name(2,background).
hnpr_object_type_name(3,card).
hnpr_object_type_name(4,button).
hnpr_object_type_name(5,edittext).
hnpr_object_type_name(6,slider).
hnpr_object_type_name(7,pulldown).
hnpr_object_type_name(8,colorselect).
hnpr_object_type_name(9,canvas).
