
#ifndef lint
static char sccsid[] = "@(#)interface.c 3.2 9/21/90 Arthur van Hoff (Turing Institute)";
#endif lint

/**********************************************************************
 * file: 	interface.c		(c)1989 Turing Institute
 * author: 	Arthur van Hoff
 *
 * Description:
 * 	A C-Interface between Quintus Prolog and HyperNeWS.
 *	All communication to and from HyperNeWS goes through the
 *	hn_any data-structure from "hypernews.h". The 'hnpr_get_*'
 *	and 'hnpr_put_*' routines build/read this data structure from
 *	prolog using the hnpr_stack. All other interface routines use this
 *	stack to communicate their arguments/results from and to prolog.
 */

#include <stdio.h>
#include "hypernews.h"

#define	OK		"ok"
#define FAIL		"fail"
#define MAXDEPTH	64

hn_any *hnpr_stack[MAXDEPTH];
hn_any **hnpr_sp = hnpr_stack;

/**********************************************************************
 * Stack operations.
 */

#define hnpr_push(any)	(*hnpr_sp++ = (any))
#define hnpr_pop()	((hnpr_sp > hnpr_stack) ? *--hnpr_sp : NULL)

void hnpr_clear(result)
char **result;
{
	hnpr_sp = hnpr_stack;
	if (hn_running)
		*result = OK;
	else
		*result = FAIL;
}

/**********************************************************************
 * Getting data from prolog onto the stack.
 */

void hnpr_put_null()
{
	hnpr_push(hn_null);
}

void hnpr_put_boolean(val)
int val;
{
	hnpr_push(hn_new_boolean(val));
}

void hnpr_put_integer(val)
int val;
{
	hnpr_push(hn_new_integer(val));
}

void hnpr_put_float(val)
float val;
{
	hnpr_push(hn_new_number(val));
}

void hnpr_put_string(val)
char *val;
{
	hnpr_push(hn_new_string(val));
}

void hnpr_put_name(val)
char *val;
{
	hnpr_push(hn_new_name(val));
}

void hnpr_put_message(name)
char *name;
{
	hn_any	*target = hnpr_pop();
	hn_any *args = hnpr_pop();

	hnpr_push(hn_new_message(target,name,args));
}

void hnpr_put_object(stack,type,name)
char *stack;
int type;
char *name;
{
	hnpr_push(hn_new_object(stack,type,name));
}

void hnpr_put_stack_object(name)
char *name;
{
	hnpr_push(hn_new_object(NULL,hn_Stack,name));
}

void hnpr_put_array(len)
int len;
{
	hnpr_push(hn_new_array_body(len));
}

void hnpr_put_array_element(n)
int n;
{
	hn_any *any = hnpr_pop();
	hn_any *array = hnpr_pop();

	array->u.array.argv[n] = any;
	hnpr_push(array);
}

/**********************************************************************
 * Getting data from the stack into prolog.
 */

void hnpr_get_integer(result,val)
char **result;
int *val;
{
	hn_any *any = hnpr_pop();

	if (any) {
		*val = any->u.number;
		hn_free(any);
		*result = OK;
	} else
		*result = FAIL;
}

void hnpr_get_float(result,val)
char **result;
float *val;
{
	hn_any *any = hnpr_pop();

	if (any) {
		*val = any->u.number;
		hn_free(any);
		*result = OK;
	} else
		*result = FAIL;
}

void hnpr_get_string(result,val)
char **result, **val;
{
	hn_any *any = hnpr_pop();

	if (any) {
		*val = any->u.string;
		hn_free(any);
		*result = OK;
	} else
		*result = FAIL;
}

void hnpr_get_name(result,val)
char **result, **val;
{
	hn_any *any = hnpr_pop();

	if (any) {
		*val = any->u.string;
		hn_free(any);
		*result = OK;
	} else
		*result = FAIL;
}

void hnpr_get_message(result,val)
char **result, **val;
{
	hn_any *any = hnpr_pop();

	if (any) {
		*val = any->u.message.name;
		hn_free(any);
		*result = OK;
	} else
		*result = FAIL;
}

void hnpr_get_array_element(result,n)
char **result;
int n;
{
	hn_any *any = hnpr_pop();

	if (any) {
		if (n < any->u.array.argc) {
			hnpr_push(any);
			hn_use(any->u.array.argv[n]);
			hnpr_push(any->u.array.argv[n]);
			*result = OK;
		}  else {
			hn_free(any);
			*result = FAIL;
		}
	} else
		*result = FAIL;
}

void hnpr_get_object(result,stack,type,name)
char **result, **stack;
int *type;
char **name;
{
	hn_any *any = hnpr_pop();

	if (any) {
		if (!(*stack = any->u.object.stack))
			*stack == "";
		*type = (int)any->u.object.type;
		*name = any->u.object.name;
		hn_free(any);
		*result = OK;
	} else
		*result = FAIL;
}

void hnpr_get_type(result,type)
char **result, **type;
{
	hn_any *any = hnpr_pop();
	int	n;

	if (any) {
		switch ((int)any->type) {
		case hn_null_type:
			*type = "null";
			break;
		case hn_number_type:
			if (any->u.number == (int)any->u.number)
				*type = "integer";
			else
				*type = "float";
			hnpr_push(any);
			break;
		case hn_boolean_type:
			if (any->u.boolean)
				*type = "true";
			else
				*type = "false";
			break;
		case hn_string_type:
			*type = "string";
			hnpr_push(any);
			break;
		case hn_name_type:
			*type = "name";
			hnpr_push(any);
			break;
		case hn_message_type:
			*type = "message";
			hn_use(any->u.message.args);
			hnpr_push(any->u.message.args);
			hn_use(any->u.message.target);
			hnpr_push(any->u.message.target);
			hnpr_push(any);
			break;
		case hn_array_type:
			*type = "array";
			hnpr_push(any);
			break;
		case hn_object_type:
			*type = "object";
			hnpr_push(any);
			break;
		default:
			*result = FAIL;
			return;
		}
		*result = OK;
	} else
		*result = FAIL;
}

/**********************************************************************
 * Here are the usual interface routines.
 */

void hnpr_read(result,delay)
char **result;
int delay;
{
	if (hn_ok()) {
		hnpr_sp = hnpr_stack;
		hnpr_push(hn_read(delay));
		*result = OK;
	} else
		*result = FAIL;
}

void hnpr_flush(result)
char **result;
{
	hn_flush();
	*result = OK;
}

void hnpr_flush_input(result)
char **result;
{
	hn_flush_input();
	*result = OK;
}

void hnpr_ok(result)
char **result;
{
	if (hn_ok())
		*result = OK;
	else
		*result = FAIL;
}

void hnpr_start(result,name)
char **result, *name;
{
	if (hn_start(name))
		*result = OK;
	else
		*result = FAIL;
}

void hnpr_stop(result)
char **result;
{
	if (hn_stop())
		*result = OK;
	else
		*result = FAIL;
}

hnpr_verbose(val)
int val;
{
	hn_verbose = val;
}

hnpr_write(result)
char **result;
{
	hn_any *any = hnpr_pop();

	if (hn_ok()) {
		hn_write(any);
		hn_flush();
		*result = OK;
	} else
		*result = FAIL;
	hn_free(any);
}

hnpr_get_any(result,param)
char **result, *param;
{
	hn_any *target = hnpr_pop();

	if (hn_ok()) {
		hnpr_push(hn_get(target,param));
		*result = OK;
	} else
		*result = FAIL;
	hn_free(target);
}

hnpr_put_any(result,param)
char **result, *param;
{
	hn_any *target = hnpr_pop();
	hn_any *value = hnpr_pop();

	if (hn_ok()) {
		if (hn_put(target,param,value)) {
			hn_flush();
			*result = OK;
		}else
			*result = FAIL;
	} else
		*result = FAIL;
	hn_free(target);
	hn_free(value);
}

hnpr_show(result)
char **result;
{
	hn_any *target = hnpr_pop();

	if (hn_ok()) {
		if (hn_show(target)) {
			hn_flush();
			*result = OK;
		} else
			*result = FAIL;
	} else
		*result = FAIL;
	hn_free(target);
}

hnpr_hide(result)
char **result;
{
	hn_any *target = hnpr_pop();

	if (hn_ok()) {
		if (hn_hide(target)) {
			hn_flush();
			*result = OK;
		} else
			*result = FAIL;
	} else
		*result = FAIL;
	hn_free(target);
}

hnpr_exists(result)
char **result;
{
	hn_any *target = hnpr_pop();

	if (hn_ok()) {
		if (hn_exists(target)) {
			*result = OK;
		} else
			*result = FAIL;
	} else
		*result = FAIL;
	hn_free(target);
}

hnpr_ps(result,postscript)
char **result, *postscript;
{
	if (hn_ok()) {
		hn_ps(" ");
		if (hn_ps(postscript)) {
			hn_ps("\n");
			hn_flush();
			*result = OK;
		} else
			*result = FAIL;
	} else
		*result = FAIL;
}

hnpr_set_drawing(result,file)
char **result, *file;
{
	hn_any *target = hnpr_pop();

	if (hn_ok()) {
		if (hn_set_drawing(target,file)) {
			hn_flush();
			*result = OK;
		} else
			*result = FAIL;
	} else
		*result = FAIL;
	hn_free(target);
}

hnpr_rename(result,name)
char **result, *name;
{
	hn_any *target = hnpr_pop(), *n;

	if (hn_ok()) {
		if (n = hn_rename(target,name)) {
			hnpr_push(n);
			*result = OK;
		} else
			*result = FAIL;
	} else
		*result = FAIL;
	hn_free(target);
}

hnpr_stack_copy(result)
char **result;
{
	hn_any *target = hnpr_pop(), *n;

	if (hn_ok()) {
		if (n = hn_stack_copy(target)) {
			hnpr_push(n);
			*result = OK;
		} else
			*result = FAIL;
	} else
		*result = FAIL;
	hn_free(target);
}
