
% "@(#)interface.pl 3.1 3/19/90 Arthur van Hoff (Turing Institute)";

%######################################################################
% file: 	interface.ps		(c)1989 Turing Institute
% author: 	Arthur van Hoff
%
% Description:
% 	Define an interface between Quintus Prolog and HyperNeWS.
%

:-
	dynamic	hnpr_verb/1.

:-
	ensure_loaded([
		'hnintf.pl'
	]).


%######################################################################
% Checks on object types etc.

hn_islist([]).
hn_islist([_|_]).

hn_object(stack(X)) :-
	atom(X).
hn_object(Stack:Object) :-
	atom(Stack),
	Object =.. [Type,Name],
	Type \== stack,
	atom(Name),
	hnpr_object_type_name(_,Type).

hn_object_type(stack(_),stack).
hn_object_type(Stack:Object,Type) :-
	atom(Stack),
	Object =.. [Type,Name],
	Type \== stack,
	atom(Name).

hn_stack(stack(X)) :-
	atom(X).
hn_card_object(Object) :-
	hn_object_type(Object,Type),
	hnpr_object_type_name(Nr,Type),
	hnpr_object_type_name(CardNr,card),
	Nr > CardNr.

hn_mystack(Stack:_,stack(Stack)) :-
	atom(Stack).

%######################################################################
% Control routines.
	
hn_start :-
	hnpr_start(ok,prolog).
hn_start(Name) :-
	atom(Name),
	hnpr_start(ok,Name).

hn_stop :-
	hnpr_stop(ok).

hn_flush :-
	hnpr_flush(ok).
hn_flush_input :-
	hnpr_flush_input(ok).

hn_ok :-
	hnpr_ok(ok).

hnpr_verb(off).

hn_verbose :-
	hn_verbose(on).

hn_verbose(on) :-
	retractall(hnpr_verb(_)),
	assert(hnpr_verb(on)),
	hnpr_verbose(1).
hn_verbose(off) :-
	retractall(hnpr_verb(_)),
	assert(hnpr_verb(off)),
	hnpr_verbose(0).

hn_clear :-
	hnpr_clear(ok).

hn_ps(PostScript) :-
	hnpr_ps(ok,PostScript).

%######################################################################
% The main event loop.

hn_run :-
	hn_clear,
	hn_read(Data),
	hn_interpret(Data), !,
	hn_run.
hn_run(TimeOut) :-
	hn_read(Data,TimeOut),
	hn_interpret(Data), !,
	hn_run(TimeOut).

hn_interpret(Data) :-
	hnpr_verb(on), 
	write('hn_read: '), write(Data), nl,
	fail.
hn_interpret(message(From,Message,Args)) :-
	predicate_property(hn_message(_,_,_),_), 
	hn_message(From,Message,Args).
hn_interpret(X) :-
	X \== message(_,_,_),
	predicate_property(hn_message(_),_), 
	hn_message(X).
hn_interpret(Data) :-
	hnpr_verb(on), 
	write('failed to interpret: '), write(Data), nl,
	fail.
hn_interpret(_).

%######################################################################
% High level I/O.

hn_read(X) :-
	hnpr_read(ok,-1),
	hnpr_get(X).

hn_read(X,Delay) :-
	hn_clear,
	hnpr_read(ok,Delay),
	hnpr_get(X).

hn_write(X) :-
	hn_clear,
	hnpr_put(X),
	hnpr_write(ok).

%######################################################################
% Sending messages.

hn_send(message(Target,Name,Args)) :-
	hn_clear,
	hnpr_put(message(Target,Name,Args)), 
	hnpr_write(ok),
	hn_flush.
hn_send(Target,Name,Args) :-
	hn_send(message(Target,Name,Args)).

%######################################################################
% Editing objects.

hn_get(Object,Arg,Value) :-
	atom(Arg), hn_object(Object), !,
	hnpr_put(Object),
	hnpr_get_any(ok,Arg),
	hnpr_get(Value).

hn_put(Object,Arg,Value) :-
	atom(Arg), hn_object(Object), !,
	hnpr_put(Value),
	hnpr_put(Object),
	hnpr_put_any(ok,Arg).

%######################################################################
% Other operations on objects

hn_show(Object) :-
	hn_clear,
	hn_object(Object), !,
	hnpr_put(Object),
	hnpr_show(ok).
hn_hide(Object) :-
	hn_clear,
	hn_object(Object), !,
	hnpr_put(Object),
	hnpr_hide(ok).
	
hn_move(Object,[X,Y]) :-
	hn_send(Object,'Move',[X,Y]).
hn_size(Object,[W,H]) :-
	hn_send(Object,'Size',[W,H]).
hn_reshape(Object,[X,Y,W,H]) :-
	hn_send(Object,'Reshape',[X,Y,W,H]).
hn_update(Object) :-
	hn_send(Object,'Update',[]).
hn_lock_update(stack(Stack)) :-
	hn_send(stack(Stack),'LockUpdate',[]).
hn_unlock_update(stack(Stack)) :-
	hn_send(stack(Stack),'UnLockUpdate',[]).

hn_connect(Object) :-
	hn_stack(Object),
	hn_send(Object,'Connect',[]).

hn_exists(Object) :-
	hn_clear,
	hn_object(Object), !,
	hnpr_put(Object),
	hnpr_exists(ok).

hn_start_edit(Object) :-
	hn_stack(Object), !,
	hn_send(Object,'StartEditMode',[]).

hn_stop_edit(Object) :-
	hn_stack(Object), !,
	hn_send(Object,'StopEditMode',[]).

hn_select(Object) :-
	hn_card_object(Object), !,
	hn_mystack(Object,Stack),
	hn_send(Stack,'SelectObject',[Object]).

hn_menu(Object,Message) :-
	hn_menu(Object,Message,[]).
hn_menu(Object,Message,Args) :-
	hn_islist(Args), hn_stack(Object), atom(Message), !,
	hn_send(Object,'OnMenu',[Args,Message]).

hn_focus_menu(Object,Message) :-
	hn_focus_menu(Object,Message,[]).
hn_focus_menu(Object,Message,Args) :-
	hn_islist(Args), hn_stack(Object), atom(Message), !,
	hn_send(Object,'OnFocusMenu',[Args,Message]).

hn_stack_menu(Object,Message) :-
	hn_stack_menu(Object,Message,[]).
hn_stack_menu(Object,Message,Args) :-
	hn_islist(Args), hn_stack(Object), atom(Message), !,
	hn_send(Object,'OnStackMenu',[Args,Message]).

hn_stack_copy(stack(Name),NewStack) :-
	hn_clear,
	hnpr_put(stack(Name)),
	hnpr_stack_copy(ok),
	hnpr_get(NewStack).

hn_rename(Object,Name,NewObject) :-
	hn_clear, \+ hn_stack(Object),
	hnpr_put(Object),
	hnpr_rename(ok,Name),
	hnpr_get(NewObject).

hn_new_object(Object,Name) :-
	atom(Name), !,
	hn_menu(Object,'NewObj',[Name]).

hn_home_card(Object) :-
	hn_stack(Object), !,
	hn_send(Object,'GoHomeCard',[]).

hn_next_card(Object) :-
	hn_stack(Object), !,
	hn_send(Object,'GoNextCard',[]).

hn_previous_card(Object) :-
	hn_stack(Object), !,
	hn_send(Object,'GoPreviousCard',[]).

hn_last_card(Object) :-
	hn_stack(Object), !,
	hn_send(Object,'GoLastCard',[]).

hn_goto_card(Object) :-
	hn_object(Object), hn_object_type(Object,card), !,
	hn_mystack(Object,Stack),
	hn_send(Stack,'GotoCard',[Object]).
hn_goto_card(Object,Nr) :-
	hn_stack(Object), integer(Nr), !,
	hn_send(Object,'GotoCard',[Nr]).
hn_goto_card(Object,Name) :-
	hn_stack(Object), atom(Name), !,
	hn_send(Object,'GotoCard',[Name]).

hn_set_drawing(Stack:button(Name),File) :-
	hn_clear,
	hnpr_put(Stack:button(Name)),
	hnpr_set_drawing(ok,File).

hn_show_error(string(Arg)) :-	
	hn_clear, atom(Arg),
	hn_write([string(Arg)]),
	hn_ps('ShowError').
hn_show_error(text(Args)) :-	
	hn_clear, hn_islist(Args),
	hn_write(text(Args)),
	hn_ps('ShowError').
hn_show_error(Args) :-	
	hn_clear, atom(Args),
	hn_write([string(Args)]),
	hn_ps('ShowError').

hn_confirm(Text,Answer) :-
	hn_connect(stack('Confirm')),
	hn_send(stack('Confirm'),'Init',[Text]),
	repeat,
	hn_read(message('Confirm':_,A,_)),
	(A = 'ConfirmYes'; A = 'ConfirmNo'), !,
	(A = 'ConfirmYes' -> Answer = yes ; Answer = no).

hn_set_text(Stack:edittext(Text),text(Value)) :-
	hn_islist(Value), !,
	hn_send(Stack:edittext(Text),'SetValue',[text(Value)]).
hn_set_text(Stack:edittext(Text),string(Value)) :-
	atom(Value), !,
	hn_send(Stack:edittext(Text),'SetValue',[[string(Value)]]).
hn_set_text(Stack:edittext(Text),Value) :-
	hn_islist(Value), !,
	hn_send(Stack:edittext(Text),'SetValue',[Value]).
hn_set_text(Stack:edittext(Text),Value) :-
	atom(Value), !,
	hn_send(Stack:edittext(Text),'SetValue',[[string(Value)]]).

hn_get_text(Stack:edittext(X),Text) :-
	hn_get(Stack:edittext(X),'Value',Text).
