; "@(#)hnintf.lisp 1.2 3/19/90 Sunil Sharma (Turing Institute)";
(setq DEBUG NIL)

;**********************************************************
; file hnintf.lisp

; Load foreign files
;**********************************************************

(load-foreign-files
 (list (format nil "~A/intf/sunlisp/interface-sun~A.o" 
	       (environment-variable "HNHOME")
	       (char (machine-type) 4)))
 (list "-lc" "-lm" (format nil "~A/bin-sun~A-xnews/libhn.a"
			   (environment-variable "HNHOME")
			   (char (machine-type) 4))))


; Low level interface between SUN Common Lisp and HyperNeWS.
;**********************************************************

  (setq *RES-ARRAY* (make-foreign-pointer 
		:type '(:pointer (:array :character (6)))))
  (setq *RES-PTR* (make-foreign-pointer 
		:type '(:pointer :character)
	        :address (foreign-pointer-address *RES-ARRAY*)))
  (defun clear-res()
	 (setf (foreign-string-value *RES-ARRAY*) " "))


; Routines for returning values to lisp.

(def-foreign-callable hnlsp_set_string
	((ptr (:pointer :character)) (str (:simple-string)))
	(setf (foreign-string-value ptr) str)
	)

(def-foreign-callable hnlsp_set_int
	((ptr (:pointer :signed-32bit)) (val :signed-32bit)) 
	(setf (foreign-value ptr) val)
	)

;**********************************************************
; Declaration of foreign functions.

(def-foreign-function (hnlsp_start
		       (:language :c)
		       (:return-type :simple-string))
  	(name :simple-string))

(def-foreign-function (hnlsp_stop
		       (:language :c)
		       (:return-type :simple-string)))

(def-foreign-function (hnlsp_ok
		       (:language :c)
		       (:return-type :simple-string)))

(def-foreign-function (hnlsp_flush
		       (:language :c)
		       (:return-type :simple-string)))

(def-foreign-function (hnlsp_flush_input
		       (:language :c)
		       (:return-type :simple-string)))

(def-foreign-function (hnlsp_verbose
		       (:language :c))
	)

(def-foreign-function (hnlsp_clear
				(:language :c)
				(:return-type :simple-string)))

(def-foreign-function (hnlsp_ps
		       (:language :c)
		       (:return-type :simple-string))
  	(postscript :simple-string))

(def-foreign-function (hnlsp_write
		       (:language :c)
		       (:return-type :simple-string)))

(def-foreign-function (hnlsp_read
			(:language :c)
		       (:return-type :simple-string))
	(delay :signed-32bit))

(def-foreign-function (hnlsp_put_any
		       (:language :c)
		       (:return-type :simple-string))
  	(param :simple-string))


(def-foreign-function (hnlsp_get_any
		       (:language :c)
		       (:return-type :simple-string))
  	(param :simple-string))

(def-foreign-function (hnlsp_show
		       (:language :c)
		       (:return-type :simple-string)))

(def-foreign-function (hnlsp_hide
		       (:language :c)
		       (:return-type :simple-string)))

(def-foreign-function (hnlsp_exists
		       (:language :c)
		       (:return-type :simple-string)))

(def-foreign-function (hnlsp_set_drawing
		       (:language :c)
		       (:return-type :simple-string))
  	(file :simple-string))

(def-foreign-function (hnlsp_rename
		       (:language :c)
		       (:return-type :simple-string))
  	(name :simple-string))

(def-foreign-function (hnlsp_stack_copy
		       (:language :c)
		       (:return-type :simple-string)))

(def-foreign-function (hnlsp_put_null
				(:language :c)))

(def-foreign-function (hnlsp_put_boolean
				(:language :c))
	(val :signed-32bit))

(def-foreign-function (hnlsp_put_integer
				(:language :c))
	(val :signed-32bit))

(def-foreign-function (hnlsp_put_float
				(:language :c))
	(val :double-float))

(def-foreign-function (hnlsp_put_string
				(:language :c))
	(val :simple-string)) 

(def-foreign-function (hnlsp_put_name
				(:language :c))
	(val :simple-string))

(def-foreign-function (hnlsp_put_message
				(:language :c))
	(name :simple-string))

(def-foreign-function (hnlsp_put_object
				(:language :c))
	(stack :simple-string)
	(type :signed-32bit)
	(name :simple-string)
	)

(def-foreign-function (hnlsp_put_stack_object
				(:language :c))
	(name :simple-string))

(def-foreign-function (hnlsp_put_array
				(:language :c))
	(len :signed-32bit)) 

(def-foreign-function (hnlsp_put_array_element
				(:language :c))
	(n :signed-32bit)) 

;*************************************************************
; Getting data from the stack into lisp.
;
(def-foreign-function (hnlsp_get_type
		       (:language :c)
		       (:return-type :simple-string))
  (result (:pointer :character)))

(def-foreign-function (hnlsp_get_integer
				(:language :c)
				(:return-type :signed-32bit))
	(result (:pointer :character)))

(def-foreign-function (hnlsp_get_float
				(:language :c)
				(:return-type :double-float))
	(result (:pointer :character)))

(def-foreign-function (hnlsp_get_string
				(:language :c)
				(:return-type :simple-string))
	(result (:pointer :character)))

(def-foreign-function (hnlsp_get_name
				(:language :c)
				(:return-type :simple-string))
	(result (:pointer :character)))

(def-foreign-function (hnlsp_get_message
				(:language :c)
				(:return-type :simple-string))
	(result (:pointer :character)))

(def-foreign-function (hnlsp_get_object
				(:language :c)
				(:return-type :simple-string))
	 (stack (:pointer :character))
	 (type (:pointer :signed-32bit))
	 (name (:pointer :character)))

(def-foreign-function (hnlsp_get_array_element
				(:language :c)
				(:return-type :simple-string))
	(n :signed-32bit))


;**************************************************************
; I/O to and from lisp.

(defun hnlsp_put (arg &optional type)
  (cond ((equal type "null") (hnlsp_put_null))
        ((equal type "true") (hnlsp_put_boolean 1))
        ((equal type "false") (hnlsp_put_boolean 0))	
        ((and (equal type "string") (stringp arg))
         (hnlsp_put_string arg))
        ((equal type "text")
	 (if (eq arg NIL) 
	     (hnlsp_put_array 0)
	     (if (listp arg) 
		 (let ()
	  	      (hnlsp_put_array (length arg))
	  	      (hnlsp_put_text arg 0)))))
        ((equal type "object")
         (let (stack1 type1 name1)
	      (multiple-value-setq (stack1 type1 name1)
				   (hnl_object_parts arg))
	      (if (equal type1 "stack")
	          (hnlsp_put_stack_object name1)
	          (if (stringp name1)
	              (hnlsp_put_object stack1 
				(hnlsp_object_type_name type1)
				name1)))))
        ((equal type "message")
         (let ((target (first arg)) (name (second arg)) 
	       (arg-array (third arg)) obj-name)
	      (setq obj-name (car (last target)))
              (if (and (listp arg-array) (hnl_object target) 
	               (stringp name))
	          (let ()
	               (hnlsp_put arg-array)
	               (hnlsp_put target "object")
	               (hnlsp_put_message name)
	               (return-from hnlsp_put t)))))
         ((eq arg NIL) (hnlsp_put_array 0))
         ((stringp arg) (hnlsp_put_string arg))
         ((integerp arg) (hnlsp_put_integer arg))
         ((floatp arg) (hnlsp_put_float arg))
         ((listp arg)
	  (let ()
	       (hnlsp_put_array (length arg))
	       (hnlsp_put_array_elements arg 0)))
	))


(defun hnlsp_put_array_elements (lst n)
  (if (null lst) (return-from hnlsp_put_array_elements))
  (hnlsp_put (car lst))
  (hnlsp_put_array_element n)
  (hnlsp_put_array_elements (cdr lst) (+ n 1)))

; assuming str-lst is a list of the text strings
(defun hnlsp_put_text (str-lst n)
  (if (null str-lst)
      (return-from hnlsp_put_text))
  (hnlsp_put (car str-lst))
  (hnlsp_put_array_element n)
  (hnlsp_put_text (cdr str-lst) (+ n 1)))





(defun hnlsp_get (type &optional arg &aux result res-array res-ptr val)
  (setq res-array (make-foreign-pointer 
		    :type '(:pointer (:array :character (6)))))
  (setq res-ptr (make-foreign-pointer 
		    :type '(:pointer :character)
		    :address (foreign-pointer-address res-array)))
  (setf (foreign-string-value res-array) " ")
  (if (null type)
      (let ()
        (setq type (hnlsp_get_type res-ptr))
        (setq result (foreign-string-value res-ptr))
        (free-foreign-pointer res-array)
	(free-foreign-pointer res-ptr)
	(if (equal result *FAIL*)
	    (return-from hnlsp_get NIL)
	    (return-from hnlsp_get (hnlsp_get type)))
	    ))

(if (equal type "message")
  (let (msg fromtype argtype from args)
       (setq msg (hnlsp_get_message res-ptr))
       (if (equal (foreign-string-value res-ptr) *FAIL*)
	   (let ()
	   (free-foreign-pointer res-array)
  	   (free-foreign-pointer res-ptr)
  	   (return-from hnlsp_get NIL)))
       (setf (foreign-string-value res-array) " ")
       (setq fromtype (hnlsp_get_type res-ptr))
       (if (equal (foreign-string-value res-ptr) *FAIL*)
	   (let ()
	   (free-foreign-pointer res-array)
  	   (free-foreign-pointer res-ptr)
  	   (return-from hnlsp_get NIL)))
       (setq from (hnlsp_get fromtype))
       (setf (foreign-string-value res-array) " ")
       (setq argstype (hnlsp_get_type res-ptr))
       (if (equal (foreign-string-value res-ptr) *FAIL*)
	   (let ()
	   (free-foreign-pointer res-array)
  	   (free-foreign-pointer res-ptr)
  	   (return-from hnlsp_get NIL)))
       (setq args (hnlsp_get argstype))
       (free-foreign-pointer res-array)
       (free-foreign-pointer res-ptr)
       (return-from hnlsp_get (list from msg args))
       ))

  (if (equal type "array")
      (return-from hnlsp_get (hnlsp_get_array 0)))
  (if (equal type "object")
      (let (obj typenr obj-type stack name 
	    stack-array stack-ptr type-ptr
	    name-array name-ptr)
           (setq stack-array (make-foreign-pointer 
	         :type '(:pointer (:array :character (30)))))
           (setq stack-ptr (make-foreign-pointer 
	         :type '(:pointer :character)
	         :address (foreign-pointer-address stack-array)))
           (setq type-ptr (make-foreign-pointer 
	         :type '(:pointer :signed-32bit)))
           (setq name-array (make-foreign-pointer 
	         :type '(:pointer (:array :character (30)))))
           (setq name-ptr (make-foreign-pointer 
	         :type '(:pointer :character)
	         :address (foreign-pointer-address name-array)))
           (setf (foreign-string-value stack-array) " ")
           (setf (foreign-string-value name-array) " ")
	   (hnlsp_get_object stack-ptr type-ptr name-ptr)
	   (setq stack (foreign-string-value stack-ptr)
		 typenr (foreign-value type-ptr)
		 name (foreign-string-value name-ptr))
	   (free-foreign-pointer stack-array)
	   (free-foreign-pointer stack-ptr)
	   (free-foreign-pointer type-ptr)
	   (free-foreign-pointer name-array)
	   (free-foreign-pointer name-ptr)
	   (setq obj-type (hnlsp_object_type_name typenr))
	   (return-from hnlsp_get (hnlsp_object_name obj-type stack name))))

  (if (equal type "true") (return-from hnlsp_get "true"))
  (if (equal type "false") (return-from hnlsp_get "false"))
  (if (equal type "null") (return-from hnlsp_get "null"))
  (if (equal type "integer")
      (setq val (hnlsp_get_integer res-ptr))
      (if (equal type "float")
	  (setq val (hnlsp_get_float res-ptr))
	  (if (equal type "string")
	      (setq val (hnlsp_get_string res-ptr))
	      (if (equal type "name")
		  (setq val (hnlsp_get_name res-ptr))))))
  (setq result (foreign-string-value res-ptr))
  (free-foreign-pointer res-array)
  (free-foreign-pointer res-ptr)
  (if (equal result *FAIL*)
      (return-from hnlsp_get NIL)
      (return-from hnlsp_get val))
  )


(defun hnlsp_get_array (n)
  (let (type (res (hnlsp_get_array_element n)))
    (if (equal res *FAIL*)
	(return-from hnlsp_get_array NIL))
    (clear-res)
    (setq type (hnlsp_get_type *RES-PTR*))
    (return-from hnlsp_get_array 
		 (append (list (hnlsp_get type)) 
			 (hnlsp_get_array (+ 1 n))))
    ))


(defun hnlsp_object_name (type stack name)
  (if (equal type "stack")
      (list "stack" name)
      (list stack (list type name))))


(defun hnlsp_object_type_name (match)
  (if (numberp match)
  (let ((nr2type '((0 "any")(1 "stack")(2 "background")
		   (3 "card")(4 "button")(5 "edittext")
		   (6 "slider")(7 "pulldown")(8 "colorselect")
		   (9 "canvas")))
	)
       (return-from hnlsp_object_type_name
		    (cadr (assoc match nr2type))))
      (if (stringp match)
          (let ((type2nr '(("any" 0)("stack" 1)("background" 2)
		   ("card" 3)("button" 4)("edittext" 5)
		   ("slider" 6)("pulldown" 7)("colorselect" 8)
		   ("canvas" 9)))
	         )
                (return-from hnlsp_object_type_name
		    (cadr (assoc match type2nr :test 'equal))))
	  (return-from hnlsp_object_type_name NIL)
	  )))

