
; "@(#)interface.lisp 1.3 3/19/90 Sunil Sharma (Turing Institute)";
;**********************************************************
; file: interface.lisp

(load (format nil "~A/intf/sunlisp/interface2.lisp"
	      (environment-variable "HNHOME")))
(load (format nil "~A/intf/sunlisp/hnintf.lisp" 
	      (environment-variable "HNHOME")))

; Low level interface between SUN Common Lisp and HyperNeWS.
;**********************************************************

(setq *OK* "ok")
(setq *FAIL* "fail")
(setq *NULL* "null")

;*********************************************************
; Checks on object types etc.

(defun hnl_object (obj &aux stack type name)
  (if (not (listp obj))
      (return-from hnl_object NIL))
  (multiple-value-setq (stack type name)
		       (hnl_object_parts obj))
  (if (equal type "stack")
      (return-from hnl_object (stringp name))
      (return-from hnl_object
		   (and (stringp stack)
			(stringp name)
			(hnlsp_object_type_name type)))))

(defun hnl_object_type (obj &aux stack type name)
  (if (not (hnl_object obj))
      (return-from hnl_object_type NIL))
  (multiple-value-setq (stack type name)
		       (hnl_object_parts obj))
  (return-from hnl_object_type type))

(defun hnl_object_name (obj &aux stack type name)
  (if (not (hnl_object obj))
      (return-from hnl_object_name NIL))
  (multiple-value-setq (stack type name)
		       (hnl_object_parts obj))
  (return-from hnl_object_name name))

(defun hnl_object_stack (object)
  (hnl_mystack object))

(defun hnl_stack (obj)
       (equal (hnl_object_type obj) "stack"))

(defun hnl_card_object (obj)
  (let (type nr card-nr)
       (setq type (hnl_object_type obj))
       (setq nr (hnlsp_object_type_name type))
       (setq card-nr (hnlsp_object_type_name "card"))
       (> nr card-nr)))

(defun hnl_mystack (object)
  (if (not (hnl_object object))
      (return-from hnl_mystack NIL))
  (if (equal (car object) "stack")
      NIL
     (list "stack" (car object))))

(defun hnl_object_parts(object &aux stack type name)
  (if (equal (car object) "stack")
      (setq stack NIL
	    type "stack"
	    name (cadr object))
      (setq stack (car object)
	    type (car (second object))
	    name (cadr (second object))))
  (values stack type name))

;**********************************************************
; Control Routines.
;**********************************************************
(defun hnl_start (&optional name)
  (if (null name)
      (setq name "lisp"))
  (equal (hnlsp_start name) *OK*))

(defun hnl_stop()
  (equal (hnlsp_stop) *OK*))

(defun hnl_flush()
  (equal (hnlsp_flush) *OK*))

(defun hnl_flush_input()
  (equal (hnlsp_flush_input) *OK*))

(defun hnl_ok()
  (equal (hnlsp_ok) *OK*))

(defun hnl_verbose(&optional state)
  (if (null state)
      (setq state 'on))
  (if (eq state 'on)
      (let ()

       (hnlsp_verbose 1))
      (let ()

       
       (hnlsp_verbose 0)))
  )

(defun hnl_clear ()
  (equal (hnlsp_clear) *OK*))

(defun hnl_ps(PostScript)
  (equal (hnlsp_ps PostScript) *OK*))

;***********************************************************
; The main event loop.

(defun hnl_run (&optional timeout)
  (do (data)
      (NIL)
      (setq data  (hnl_read timeout))
      (if (null (hnl_interpret data))
	  (return-from hnl_run NIL))
      ))

(defun hnl_interpret (data &aux from message args)
  (if (or (null data) (equal data "timeout"))
      (return-from hnl_interpret NIL))
  (setq from (car data)
	message (second data)
	args (third data))
  (hnl_call_handler from message args)
  (return-from hnl_interpret T)
  )


;*******************************************************************
; High level I/O
;*******************************************************************

(defun hnl_read(&optional delay)
  (if (null delay)
      (setq delay -1)
      (hnl_clear))
  (if (equal (hnlsp_read delay) *FAIL*)
      (return-from hnl_read NIL))
  (hnlsp_get NIL)
  )


(defun hnl_write(arg &optional type)
  (hnl_clear)
  (hnlsp_put arg type)
  (equal (hnlsp_write) *OK*))

;*******************************************************************
; Sending Messages.
;*******************************************************************

(defun hnl_send (message &optional arg2 arg3)
  (if (hnl_object message)
      (setq message (list message arg2 arg3)))
  (hnl_clear)
  (hnlsp_put message "message")
  (if (equal (hnlsp_write) *FAIL*)
      (return-from hnl_send NIL)
      (hnl_flush)))

;*****************************************************************
; Editing objects.

(defun hnl_get(object arg)
  (if (not (and (stringp arg)
		(hnl_object object)))
      (return-from hnl_get NIL))
  (hnlsp_put object "object")
  (if (equal (hnlsp_get_any arg) *FAIL*)
      (return-from hnl_get NIL))
  (return-from hnl_get (hnlsp_get NIL)))

(defun hnl_put(object arg value)
  (if (not (and (stringp arg)
		(hnl_object object)))
      (return-from hnl_put NIL))
  (hnlsp_put value)
  (hnlsp_put object "object")
  (equal (hnlsp_put_any arg) *OK*))


;****************************************************************
; Other operation on objects.

(defun hnl_show (object)
  (if (not (hnl_clear)) (return-from hnl_show NIL))
  (if (not (hnl_object object)) (return-from hnl_show NIL))
  (hnlsp_put object "object")
  (equal (hnlsp_show) *OK*))

(defun hnl_hide (object)
  (if (not (hnl_clear)) (return-from hnl_hide NIL))
  (if (not (hnl_object object)) (return-from hnl_hide NIL))
  (hnlsp_put object "object")
  (equal (hnlsp_hide) *OK*))

(defun hnl_move (object coord-list)
  (hnl_send object "Move" coord-list))

(defun hnl_size (object wh-list)
  (hnl_send object "Size" wh-list))

(defun hnl_reshape (object xywh-list)
  (hnl_send object "Reshape" xywh-list))

(defun hnl_update (object)
  (hnl_send object "Update" NIL))

(defun hnl_lock_update (stack)
  (if (not (hnl_stack stack))
      (return-from hnl_lock_update NIL))
  (hnl_send stack "LockUpdate" NIL))

(defun hnl_unlock_update (stack)
  (if (not (hnl_stack stack))
      (return-from hnl_unlock_update NIL))
  (hnl_send stack "UnLockUpdate" NIL))

(defun hnl_connect(stack)
  (if (not (hnl_stack stack))
      (setq stack (list "stack" stack)) )
  (if (not (hnl_stack stack))
      (return-from hnl_connect NIL))
  (hnl_send stack "Connect" NIL))

(defun hnl_exists (object)
  (if (not (hnl_clear)) (return-from hnl_exists NIL))
  (if (not (hnl_object object)) (return-from hnl_exists NIL))
  (hnlsp_put object "object")
  (equal (hnlsp_exists) *OK*))

(defun hnl_start_edit (object)
  (if (not (hnl_stack object))
      (return-from hnl_start_edit NIL))
  (hnl_send object "StartEditMode" NIL))

(defun hnl_stop_edit (object)
  (if (not (hnl_stack object))
      (return-from hnl_stop_edit NIL))
  (hnl_send object "StopEditMode" NIL))

(defun hnl_select (object)
  (if (not (hnl_card_object object))
      (return-from hnl_select NIL))
  (hnl_send (hnl_mystack object) "SelectObject" (list object)))

(defun hnl_menu (object message &optional args)
  (if (not (and (listp args) (hnl_stack object)
		(stringp message)))
      (return-from hnl_menu NIL))
  (hnl_send object "OnMenu" (list args message)))

(defun hnl_focus_menu (object message &optional args)
  (if (not (and (listp args) (hnl_stack object)
		(stringp message)))
      (return-from hnl_focus_menu NIL))
  (hnl_send object "OnFocusMenu" (list args message)))

(defun hnl_stack_menu (object message &optional args)
  (if (not (and (listp args) (hnl_stack object)
		(stringp message)))
      (return-from hnl_stack_menu NIL))
  (hnl_send object "OnStackMenu" (list args message)))

(defun hnl_stack_copy (stack)
  (if (not (hnl_stack stack))
      (return-from hnl_stack_copy NIL))
  (hnl_clear)
  (hnlsp_put stack "object")
  (if (not (equal (hnlsp_stack_copy) *OK*))
      (return-from hnl_stack_copy NIL))
  (return-from hnl_stack_copy (hnlsp_get NIL)))

(defun hnl_rename (object name)
  (hnl_clear)
  (if (hnl_stack object)
      (return-from hnl_rename NIL))
  (hnlsp_put object "object")
  (if (not (equal (hnlsp_rename name) *OK*))
      (return-from hnl_rename NIL))
  (return-from hnl_rename (hnlsp_get NIL)))

(defun hnl_new_object (object name)
  (if (not (stringp name))
      (return-from hnl_new_object NIL))
  (hnl_menu object "NewObj" (list name)))

(defun hnl_home_card (object)
  (if (not (hnl_stack object))
      (return-from hnl_home_card NIL))
  (hnl_send object "GoHomeCard" NIL))

(defun hnl_next_card (object)
  (if (not (hnl_stack object))
      (return-from hnl_next_card NIL))
  (hnl_send object "GoNextCard" NIL))

(defun hnl_previous_card (object)
  (if (not (hnl_stack object))
      (return-from hnl_previous_card NIL))
  (hnl_send object "GoPreviousCard" NIL))

(defun hnl_last_card (object)
  (if (not (hnl_stack object))
      (return-from hnl_last_card NIL))
  (hnl_send object "GoLastCard" NIL))

(defun hnl_goto_card (object &optional arg)
  (if (equal (hnl_object_type object) "card")
      (setq arg NIL))
  (cond ((null arg)
	 (if (not (and
		   (hnl_object object)
		   (equal (hnl_object_type object) "card")))
	     (return-from hnl_goto_card NIL)
           (hnl_goto_card (hnl_mystack object) (hnl_object_name object))))
	((not (hnl_stack object))
	 (return-from hnl_goto_card NIL))
	((or (integerp arg) (stringp arg))
	 (hnl_send object "GotoCard" (list arg)))
	(T (return-from hnl_goto_card NIL))
	))


(defun hnl_set_drawing (butt-obj file)
  (hnl_clear)
  (if (not (equal (hnl_object_type butt-obj) "button"))
      (return-from hnl_set_drawing NIL))
  (hnlsp_put butt-obj "object")
  (return-from hnl_set_drawing (hnlsp_set_drawing file)))

(defun hnl_show_error (arg)
  (hnl_clear)
  (cond ((stringp arg)
	 (hnl_write arg "string"))
	((listp arg)
	 (hnl_write arg "text")))
  (hnl_ps "ShowError"))

(defun hnl_confirm (text &aux answer)
  (hnl_connect "Confirm")
  (hnl_send '("stack" "Confirm") "Init" (list text))
  (do ()
      ((or (equal answer "ConfirmYes") 
	   (equal answer "ConfirmNo"))
       (equal answer "ConfirmYes"))
      (setq answer (second (hnl_read)))))

(defun hnl_set_text (text-object value)
  (if (not (equal (hnl_object_type text-object) "edittext"))
      (return-from hnl_set_text NIL))
  (setq value (list value))
  (hnl_send text-object "SetValue" value))

(defun hnl_get_text (text-object)
  (hnl_get text-object "Value")
  )

(defun hnl_save (stack)
  (if (not (hnl_stack stack))
      (return-from hnl_save NIL))
  (hnl_send stack "Save"))

(defun hnl_save_as(stack)
  (if (not (hnl_stack stack))
      (return-from hnl_save_as nil))
  (hnl_send stack "SaveAs"))


;***********************************************************

(defun hnl_message (message &aux object name args)
  (setq object (first message)
	name (second message)
	args (third message))
  (values object name args))

(if (null (foreign-undefined-symbol-names))
    (format t "All foreign Symbols are now defined~%"))

