; "@(#)interface2.lisp 1.2 3/19/90 Sunil Sharma (Turing Institute)";
; File: interface2.lisp.
; Higher level functions for the lisp-hypernews interface.

;**** Define a global variable where interests are registered
(defvar *hnl_interests* nil)

(defun hnl_clear_interests ()
  (setq *hnl_interests* nil))

;*********************************************************
; Register objects used in the interface.
; When declaring stacks, the first argument is NIL.

(defun hnl_interface_object ( stack type name)
  (if (or (and stack (symbolp stack)) 
	  (symbolp name))
      (return-from hnl_interface_object NIL))
  (if (symbolp type )
      (setq type (string-downcase (symbol-name type))))
  (cond ((equal type "stack")
         (return-from hnl_interface_object
		      (list "stack" name)))
	(t (if (listp stack)
	       (setq stack (cadr stack)))
           (return-from hnl_interface_object 
			(list stack (list type name))))
	))

;**********************************************************
; Specify the handler, "func-sym", to be called when a message
; of name "name" arrives from the object "from".

(defun hnl_set_handler (from message func-sym)
  (if (stringp func-sym)
      (setq func-sym (intern (string-upcase func-sym))))
  (setq *hnl_interests* 
	(append *hnl_interests* 
		(list (list (list from message) func-sym)))))


;**********************************************************
; Call the handler, if any, for a message.

(defun hnl_call_handler (from message &optional args)
  (let ((handler (hnl_get_handler from message)))
       (setq args (list* from message (if args args
					  (list NIL))))
    (if handler (apply handler (list args)))))

(defun hnl_get_handler (from message &aux handler-name item)
  (dolist (interest *hnl_interests*)
	  (setq item (car interest))
	  (if (and (or (null (car item))
		       (equal (car item) from))
		   (or (null (second item))
		       (equal (second item) message)))
	      (return-from hnl_get_handler (cadr interest))))
  (return-from hnl_get_handler NIL))

;**********************************************************
; General routines for setting and getting object values

(defun hnl_set_value (target value)
  (if (not (listp value))
      (setq value (list value)))
  (if (equal (hnl_object_type target) "edittext")
      (hnl_set_text target value)
    (hnl_send target "SetValue" value)))

(defun hnl_get_value (target)
  (hnl_get target "Value"))

