
#ifndef lint
static char sccsid[] = "@(#)obj.c 1.3 7/19/89 Arthur van Hoff (Turing Institute)";
#endif lint


#include <stdio.h>
#include "psdata.h"

#define	MINSTACK	1000
#define MAXDSTACK	20

int exec_level = 0, initialized = 0;

struct psobject *currentfile = NULL;

int	stacksize = 0;
struct psobject	**psstack = NULL;
struct psobject **stack = NULL;
struct psobject *dstack[MAXDSTACK], **dstackptr = dstack;
struct psobject *currentdict, *systemdict;

int	trace = 0, breakflag = 0;
struct  psobject *first_free = NULL;

struct psobject *
new_object(type)
{
	struct psobject *obj = first_free;

	if (obj == NULL) {
		obj = (struct psobject *)malloc(sizeof(struct psobject));
		if (obj == NULL)
			error("out of memory","");
	} else 
		first_free = (struct psobject *)first_free->val.str;
	obj->type = type;
	obj->exec = FALSE;
	obj->ref = 0;
	return(obj);
}

use_object(obj)
struct psobject *obj;
{
	obj->ref++;
}

unuse_object(obj)
struct psobject *obj;
{
	struct keyvalpair *pair;
	int	n;

	if (!obj->ref)
		error("ERROR: trying to unuse dead object\n","");

	if (!(--obj->ref)) {
		/*
		printf("destroy(");
		print_object(stdout,obj);
		printf(")\n");
		*/
		switch (obj->type) {
			case DICT:
				pair = obj->val.dictdef->first;
				while (pair != NULL) {
					unuse_object(pair->key);
					unuse_object(pair->val);
					pair = pair->next;
				}
				free(obj->val.dictdef->pair);
				free(obj->val.dictdef);
				break;
			case ARRAY:
				n = obj->val.arr->len;
				while (n--)
					if (obj->val.arr->val[n] != NULL)
					    unuse_object(obj->val.arr->val[n]);
				free(obj->val.arr);
				break;
			case STRING:
			case NAME:
				free(obj->val.str);
				break;
			case STREAM:
				free(obj->val.file->name);
				fclose(obj->val.file->fp);
				free(obj->val.file);
				break;
		}
		obj->val.str = (char *)first_free;
		first_free = obj;
	}
}

equal_object(obj1,obj2)
struct psobject *obj1, *obj2;
{
	if (obj1->type != obj2->type) 
		return(FALSE);
	if (obj1 == obj2)
		return(TRUE);
	switch (obj1->type) {
	case FLOAT:
		return(obj1->val.floatval == obj2->val.floatval);
	case BOOL:
		return(obj1->val.boolval == obj2->val.boolval);
	case MARK:
	case NILL:
		return(TRUE);
	case OPERATOR:
		return(obj1->val.proc == obj2->val.proc);
	case NAME:
	case STRING:
		return(!strcmp(obj1->val.str,obj2->val.str));
	default:
		return(FALSE);
	}
}

/* Escapes '(' and ')' in strings (TN) */
print_escaped_string(fp, str)
FILE *fp;
register char *str;
{
	putc('(', fp);
	while (*str) {
		if ((*str == '(') || (*str == ')')) 
			fputc('\\', fp);
		putc((*str++), fp);
	}
	putc(')', fp);
	return(0);
}

print_object(fp,obj)
FILE *fp;
struct psobject *obj;
{
	float n;

	switch (obj->type) {
	case DICT:
		fprintf(fp,"dict[%d]",obj->val.dictdef->maxlen);
		break;
	case ARRAY:
		fprintf(fp,(obj->exec ? "array{%d}" : "array[%d]"),
			obj->val.arr->len);
		break;
	case FLOAT:
		n = (int)obj->val.floatval;
		if (obj->val.floatval == n)
			fprintf(fp,"%d",(int)obj->val.floatval);
		else
			fprintf(fp,"%.3f",obj->val.floatval);
		break;
	case NAME:
		fprintf(fp,(obj->exec ? "%s" : "/%s"),obj->val.str);
		break;
	case STRING:
		print_escaped_string(fp, obj->val.str);
		break;
	case MARK:
		fprintf(fp,"mark");
		break;
	case BOOL:
		fprintf(fp,(obj->val.boolval ? "true" : "false"));
		break;
	case NILL:
		fprintf(fp,"null");
		break;
	case STREAM:
		fprintf(fp,(obj->exec ? "file{%d,%d,'%s'}" : "file[%d,%d,'%s']"),
			obj->val.file->ln,
			obj->val.file->col,
			obj->val.file->name);
		break;
	case OPERATOR:
		fprintf(fp,"<operator>");
		break;
	default:
		fprintf(fp,"<UNKNOWN>");
	}
}

output_object(fp,obj)
FILE *fp;
struct psobject *obj;
{
	float n;
	register int i;
	register struct keyvalpair *p;

	switch (obj->type) {
	case DICT:
		fprintf(fp,"%d dict begin\n",obj->val.dictdef->maxlen);
		for (p = obj->val.dictdef->first ; p ; p = p->next) {
			output_object(fp,p->key);
			output_object(fp,p->val);
			fprintf(fp,"def\n");
		}
		fprintf(fp,"currentdict end\n",obj->val.dictdef->maxlen);
		break;
	case ARRAY:
		fprintf(fp,(obj->exec ? "{" : "["));
		for (i = 0 ; i < obj->val.arr->len ; i++)
			output_object(fp,obj->val.arr->val[i]);
		fprintf(fp,(obj->exec ? "}\n" : "]\n"));
		break;
	case FLOAT:
		n = (int)obj->val.floatval;
		if (obj->val.floatval == n)
			fprintf(fp,"%d ",(int)obj->val.floatval);
		else
			fprintf(fp,"%.3f ",obj->val.floatval);
		break;
	case NAME:
		fprintf(fp,(obj->exec ? "%s\n" : "/%s "),obj->val.str);
		break;
	case STRING:
		print_escaped_string(fp, obj->val.str);
		break;
	case MARK:
		fprintf(fp,"mark ");
		break;
	case BOOL:
		fprintf(fp,(obj->val.boolval ? "true " : "false "));
		break;
	case NILL:
		fprintf(fp,"null ");
		break;
	case STREAM:
		fprintf(fp,"null ");
		break;
	case OPERATOR:
		fprintf(fp,"<operator>");
		break;
	default:
		fprintf(fp,"<UNKNOWN>");
	}
}

exec_object(obj)
struct psobject *obj;
{
	struct keyvalpair *pair;
	int	n;

	if (obj == NULL)
		return;

	if (trace) {
		printf("EXEC(");
		print_object(stdout,obj);
		printf(")\n");
	}
	if (obj->exec)
		switch (obj->type) {
			case NAME:
				pair = dict_search(obj);
				if (pair == NULL)
					error("undefined:",obj->val.str);
				if (pair->val->exec) {
					use_object(pair->val);
					exec_object(pair->val);
					unuse_object(pair->val);
					return;
				} 
				use_object(pair->val);
				push(pair->val);
				return;
			case ARRAY:
				n = 0;
				while (!(breakflag) && (n < obj->val.arr->len)) 
					if (obj->val.arr->val[n]->type == ARRAY) {
						use_object(obj->val.arr->val[n]);
						push(obj->val.arr->val[n++]);
					} else {
						use_object(obj->val.arr->val[n]);
						exec_object(obj->val.arr->val[n]);
						unuse_object(obj->val.arr->val[n++]);
					}
				return;
			case OPERATOR:
				obj->val.proc();
				return;
			case STREAM:
				exec_file(obj);
				return;
		}
	use_object(obj);
	push(obj);
}

char *
stralloc(str)
char *str;
{
	char *s = (char *)malloc(strlen(str)+1);
	strcpy(s,str);
	return(s);
}

error(str1,str2)
char *str1, *str2;
{
	fprintf(stderr,"error: %s %s\n",str1,str2);
	exit(-1);
}

read_char_from_file()
{
	int c = getc(currentfile->val.file->fp);

	if (c == '\n') {
		currentfile->val.file->col = 0;
		currentfile->val.file->ln++;
	} else 
		currentfile->val.file->col++;
	return(c);
}

exec_file(file)
struct psobject *file;
{
	struct psobject *obj, *oldfile = currentfile;

	use_object(currentfile = file);
	
	while ((obj = get_token()) != NULL) {
		if (obj->exec && (obj->type == NAME)) {
			if (!strcmp("{",obj->val.str)) {
				use_object(obj);
				exec_object(obj);
				exec_level++;
				unuse_object(obj);
			} else if (!strcmp("}",obj->val.str)) {
				use_object(obj);
				exec_object(obj);
				if (exec_level)
					exec_level--;
				unuse_object(obj);
			} else if (exec_level) {
				use_object(obj);
				push(obj);
			} else {
				use_object(obj);
				exec_object(obj);
				unuse_object(obj);
			}
		} else {
			use_object(obj);
			push(obj);
		}
	}

	unuse_object(currentfile);
	currentfile = oldfile;
}

init_all()
{
	FILE *fp;
	char fname[256];

	if (initialized)
		return;
	
	strncpy(fname,getenv("HNHOME"),256);
	strncat(fname,INITFILE,256);

	if ((fp = fopen(fname,"r")) == NULL)
		error("could not open:",fname);

	currentdict = systemdict = new_dict(100);
	use_object(currentdict);
	init_ops();
	exec_fp(fp,INITFILE);
	initialized = 1;
}

exec_fp(fp,name)
FILE *fp;
char *name;
{
	struct psobject *file_obj = new_object(STREAM);

	file_obj->val.file = (struct file_def *)malloc(sizeof(struct file_def));
	file_obj->val.file->fp = fp;
	file_obj->val.file->name = stralloc(name);
	file_obj->val.file->ln = 0;
	file_obj->val.file->col = 0;
	file_obj->exec = TRUE;
	use_object(file_obj);
	exec_file(file_obj);
	unuse_object(file_obj);
}

/******************************************************
	DICTIONARIES
******************************************************/

dict_hash_val(obj)
struct psobject *obj;
{
	int	n;

	switch (obj->type) {
		case OPERATOR:
		case STREAM:
		case DICT:	
		case ARRAY:	
			return((int)obj);
		case NAME:
		case STRING:	
			n = strlen(obj->val.str);
			return((n * obj->val.str[0])+obj->val.str[n-1]);
		case FLOAT:
			return((int)obj->val.floatval*1235);
		case BOOL:
			return(obj->val.boolval);
		case NILL:
		case MARK:
			return(0);
	}
}

struct psobject *
new_dict(size)
{
	struct psobject *obj = new_object(DICT);
	obj->val.dictdef = (struct dictdef *)malloc(sizeof(struct dictdef));
	obj->val.dictdef->maxlen = size;
	obj->val.dictdef->pair = (struct keyvalpair *)malloc(
			sizeof(struct keyvalpair) * size);
	obj->val.dictdef->first = NULL;
	obj->val.dictdef->pair[0].hash = NULL;
	obj->val.dictdef->free = obj->val.dictdef->pair;
	while (--size) {
		obj->val.dictdef->pair[size].hash = NULL;
		obj->val.dictdef->pair[size-1].next = 
			&obj->val.dictdef->pair[size];
	}
	return(obj);
}

dict_begin(dict)
struct psobject *dict;
{
	if (dstackptr >= &dstack[MAXDSTACK])
		error("dict stack overflow","");
	*dstackptr++ = currentdict;
	currentdict = dict;
	use_object(currentdict);
}

dict_end()
{
	if (dstackptr <= dstack)
		error("dict stack underflow","");
	unuse_object(currentdict);
	currentdict = *--dstackptr;
}

struct keyvalpair *
dict_lookup(dict,key)
struct psobject *dict, *key;
{
	struct keyvalpair *p = dict->val.dictdef->pair[dict_hash_val(key) %
					dict->val.dictdef->maxlen].hash;

	while (p != NULL) {
		if (equal_object(p->key,key))
			return(p);
		p = p->coll;
	}
	return(NULL);
}

struct psobject *
dict_where(key)
struct psobject *key;
{
	struct keyvalpair *p;
	struct psobject **dp = dstackptr;
	
	if ((p = dict_lookup(currentdict,key)) != NULL)
		return(currentdict);
	while (dp-- > dstack)
		if ((p = dict_lookup(*dp,key)) != NULL)
			return(*dp);

	return(NULL);
}

struct keyvalpair *
dict_search(key)
struct psobject *key;
{
	struct keyvalpair *p;
	struct psobject **dp = dstackptr;
	
	if ((p = dict_lookup(currentdict,key)) != NULL)
		return(p);
	while (dp > dstack)
		if ((p = dict_lookup(*--dp,key)) != NULL)
			return(p);

	return(NULL);
}

dict_define(dict,key,val)
struct psobject *dict, *key, *val;
{
	struct keyvalpair *p = dict_lookup(dict,key);
	int	n;

	if (p != NULL) {
		if (trace) {
			printf("redefine ");
			print_object(stdout,key);
			printf(" to be ");
			print_object(stdout,val);
			printf(" was ");
			print_object(stdout,p->val);
			printf("\n");
		}
		unuse_object(p->val);
		use_object(p->val = val);
	} else {
		if (trace) {
			printf("define ");
			print_object(stdout,key);
			printf(" to be ");
			print_object(stdout,val);
			printf("\n");
		}
		if ((p = dict->val.dictdef->free) == NULL)
			error("def","dictfull");
		dict->val.dictdef->free = p->next;
		use_object(p->key = key);
		use_object(p->val = val);
		p->next = dict->val.dictdef->first;
		dict->val.dictdef->first = p;
		/* insert it in the collision list */
		n = (dict_hash_val(key) % dict->val.dictdef->maxlen);
		p->coll = dict->val.dictdef->pair[n].hash;
		dict->val.dictdef->pair[n].hash = p;
	}
}

define_operator(name,proc)
char *name;
int (*proc)();
{
	struct psobject *obj1 = new_object(NAME);	
	struct psobject *obj2 = new_object(OPERATOR);	

	obj1->val.str = stralloc(name);
	obj2->val.proc = proc;
	obj2->exec = TRUE;

	dict_define(systemdict,obj1,obj2);
}


/******************************************************
	STACK
******************************************************/


push(obj)
struct psobject *obj;
{
	if (!psstack) {
		psstack = (struct psobject **)
			malloc(MINSTACK * sizeof(struct psobject *));
		stacksize = MINSTACK;
		stack = psstack;
	}
	if ((stack - psstack) >= stacksize) {
		struct psobject **newstack = (struct psobject **)
				malloc((stacksize  + MINSTACK) * 
					sizeof(struct psobject *));
		if (!newstack)
			error("stack overflow","");
		stacksize += MINSTACK;
		memcpy(newstack,psstack,stacksize * sizeof(struct psobject *));
		stack = newstack + (stack - psstack);
		free(psstack);
		psstack = newstack;
	}
	*stack++ = obj;
}

struct psobject *
pop()
{
	if (stack <= psstack)
		error("stack underflow","");
	return(*--stack);
}

