
#ifndef lint
static char sccsid[] = "@(#)ops.c 3.1 3/19/90 Arthur van Hoff (Turing Institute)";
#endif lint


#include <stdio.h>
#include "psdata.h"

#define	R2D(x)	((double)((x * 6.28319)/360.0))

double 	tan(), sin(), cos(), floor(), sqrt(), log();

#define	ECHO(x)x
#define BINOP(op,stat,type)\
ECHO(op)()\
{\
	struct psobject *obj1 = pop_object("op",type);\
	struct psobject *obj2 = pop_object("op",type);\
	struct psobject *obj3 = new_object(type);\
	stat;\
	use_object(obj3); push(obj3); unuse_object(obj1); unuse_object(obj2);\
}
#define UNOP(op,stat,type)\
ECHO(op)()\
{\
	struct psobject *obj1 = pop_object("op",type);\
	struct psobject *obj2 = new_object(type);\
	stat;\
	use_object(obj2); push(obj2); unuse_object(obj1);\
}


struct psobject *true_object, *null_object, *mark_object;

struct psobject *
pop_object(str,type)
char *str;
{
	struct psobject *obj = pop();

	if (obj->type != type)
		error(str,"type check");
	return(obj);
}

null()
{
	use_object(null_object);
	push(null_object);
}

true()
{
	use_object(true_object);
	push(true_object);
}

lt()
{
	struct psobject *obj1 = pop();
	struct psobject *obj2 = pop();
	struct psobject *obj3 = new_object(BOOL);

	if (obj1->type != obj2->type)
		error("lt","type check");
	if (obj1->type == FLOAT)
		obj3->val.boolval = (obj2->val.floatval < obj1->val.floatval);
	else if (obj1->type == STRING)
		obj3->val.boolval = (strcmp(obj2->val.str,obj1->val.str) < 0);
	else obj3->val.boolval = FALSE;

	use_object(obj3);
	push(obj3);
	unuse_object(obj1);
	unuse_object(obj2);
}

eq()
{
	struct psobject *obj1 = pop();
	struct psobject *obj2 = pop();
	struct psobject *obj3 = new_object(BOOL);

	obj3->val.boolval = equal_object(obj1,obj2);
	use_object(obj3); push(obj3); unuse_object(obj1); unuse_object(obj2);
}

BINOP(nand,obj3->val.boolval = !(obj1->val.boolval && obj2->val.boolval),BOOL);
BINOP(mul,obj3->val.floatval = obj2->val.floatval * obj1->val.floatval,FLOAT);
BINOP(sub,obj3->val.floatval = obj2->val.floatval - obj1->val.floatval,FLOAT);
BINOP(div,obj3->val.floatval = obj2->val.floatval / obj1->val.floatval,FLOAT);
BINOP(bitshift,obj3->val.floatval = ((int)obj2->val.floatval << (int)obj1->val.floatval),FLOAT);
UNOP(sqrt_op,obj2->val.floatval = sqrt(obj1->val.floatval),FLOAT);
UNOP(floor_op,obj2->val.floatval = floor(obj1->val.floatval),FLOAT);
UNOP(sin_op,obj2->val.floatval = sin(R2D(obj1->val.floatval)),FLOAT);
UNOP(cos_op,obj2->val.floatval = cos(R2D(obj1->val.floatval)),FLOAT);
UNOP(tan_op,obj2->val.floatval = tan(R2D(obj1->val.floatval)),FLOAT);
UNOP(ln,obj2->val.floatval = log(obj1->val.floatval),FLOAT);

pop_op()
{
	unuse_object(pop());
}

clear()
{
	while (stack > psstack)
		unuse_object(pop());
}

exch()
{
	struct psobject *obj1 = pop();
	struct psobject *obj2 = pop();
	push(obj1);
	push(obj2);
}

index()
{
	struct psobject *obj1 = pop_object("index",FLOAT);
	int n1 = stack - psstack, n2 = obj1->val.floatval;

	if (n1 <= n2)
		error("index:","stack underflow");
	use_object(*(stack - n2 - 1));
	push(*(stack - n2 - 1));
	unuse_object(obj1);
}

mark()
{
	use_object(mark_object);
	push(mark_object);
}

counttomark()
{
	struct psobject **p = stack, *obj;
	int	 n = 0;

	while (p > psstack) {
		if ((*--p)->type == MARK)
			break;
		n++;
	}
	obj = new_object(FLOAT);
	obj->val.floatval = n;
	use_object(obj);
	push(obj);
}

string()
{
	struct psobject *obj = pop_object("string",FLOAT),
			*strobj = new_object(STRING);
	int n = obj->val.floatval;
	char	*str = (char *)malloc(n+1);

	if (str == NULL)
		error("out of space");
	while (n--) 
		str[n] = '\001';

	strobj->val.str = stralloc(str);
	if (strobj->val.str == NULL)
		error("string:","out of memory");
	free(str);
	use_object(strobj);
	push(strobj);
	unuse_object(obj);
}

array()
{
	struct psobject *obj = pop_object("array",FLOAT),
			*arr = new_object(ARRAY);
	int n = obj->val.floatval;

	arr->val.arr = (struct array_def *)malloc(sizeof(short) + 
		(sizeof(struct psobject *) * n));
	if (arr->val.arr == NULL)
		error("array:","out of memory");
	arr->val.arr->len = n;
	while (n--) 
		arr->val.arr->val[n] = NULL;
	use_object(arr);
	push(arr);
	unuse_object(obj);
}

type_op()
{
	struct psobject *obj = pop();
	struct psobject *res = new_object(NAME);

	switch(obj->type) {
		case DICT: 	res->val.str = stralloc("dicttype"); break;
		case ARRAY: 	res->val.str = stralloc("arraytype"); break;
		case STRING: 	res->val.str = stralloc("stringtype"); break;
		case NAME: 	res->val.str = stralloc("nametype"); break;
		case FLOAT: 	res->val.str = stralloc("realtype"); break;
		case BOOL: 	res->val.str = stralloc("booleantype"); break;
		case OPERATOR: 	res->val.str = stralloc("operatortype"); break;
		case NILL: 	res->val.str = stralloc("nulltype"); break;
		case MARK: 	res->val.str = stralloc("marktype"); break;
		case STREAM: 	res->val.str = stralloc("filetype"); break;
		default:	res->val.str = stralloc("unkowntype"); break;
	}
	res->exec = TRUE;
	use_object(res);
	push(res);
	unuse_object(obj);
}

length()
{
	struct psobject *obj = pop();
	struct psobject *res;
	struct keyvalpair *pair;
	int	n = 0;

	if (obj->type == DICT) {
		pair = obj->val.dictdef->first;
		while(pair != NULL) {
			pair = pair->next;
			n++;
		}
	} else if (obj->type == ARRAY)
		n = obj->val.arr->len;
	else if (obj->type == STRING)
		n = strlen(obj->val.str);
	res = new_object(FLOAT);
	res->val.floatval = n;
	use_object(res);
	push(res);
	unuse_object(obj);
}

astore()
{
	struct psobject *arr = pop_object("astore",ARRAY);
	int n = max(0,arr->val.arr->len);
	
	while (n--)
		arr->val.arr->val[n] = pop();
	push(arr);
}

aload()
{
	struct psobject *arr = pop_object("aload",ARRAY);
	int n = 0;

	while (n < arr->val.arr->len) {
		use_object(arr->val.arr->val[n]);
		push(arr->val.arr->val[n]);
		n++;
	}
	push(arr);
}

cvx()
{
	struct psobject *obj = pop();
	obj->exec = TRUE;
	push(obj);
}

exec()
{
	struct psobject *obj = pop();

	exec_object(obj);
	unuse_object(obj);
}

ifelse()
{
	struct psobject *obj1 = pop();
	struct psobject *obj2 = pop();
	struct psobject *obj3 = pop_object("ifelse",BOOL);

	if (obj3->val.boolval)
		exec_object(obj2);
	else
		exec_object(obj1);

	unuse_object(obj1);
	unuse_object(obj2);
	unuse_object(obj3);
}

for_op()
{
	struct psobject *obj1 = pop();
	struct psobject *obj2 = pop_object("for",FLOAT);
	struct psobject *obj3 = pop_object("for",FLOAT);
	struct psobject *obj4 = pop_object("for",FLOAT);
	float	from = obj4->val.floatval, 
		to = obj2->val.floatval, 
		step = obj3->val.floatval;

	unuse_object(obj2);
	unuse_object(obj3);
	unuse_object(obj4);
	
	while ((!breakflag) && (from <= to)) {
		obj2 = new_object(FLOAT);
		obj2->val.floatval = from;
		use_object(obj2);
		push(obj2);
		exec_object(obj1);
		from += step;
	}
	unuse_object(obj1);
	breakflag = 0;
}

forall_op()
{
	struct psobject *obj1 = pop();
	struct psobject *obj2 = pop();
	struct psobject *obj3;
	struct keyvalpair *pair;
	int	n;

	if (obj2->type == DICT) {
		pair = obj2->val.dictdef->first;
		while (pair != NULL) {
			use_object(pair->key);
			use_object(pair->val);
			push(pair->key);
			push(pair->val);
			exec_object(obj1);
			pair = pair->next;
		}
	} else if (obj2->type == ARRAY) {
		for (n = 0 ; n < obj2->val.arr->len ; n++) {
			use_object(obj2->val.arr->val[n]);
			push(obj2->val.arr->val[n]);
			exec_object(obj1);
		}
	} else if (obj2->type == STRING) {
		for (n = 0 ; n < strlen(obj2->val.str) ; n++) {
			obj3 = new_object(FLOAT);
			obj3->val.floatval = obj2->val.str[n];
			use_object(obj3);
			push(obj3);
			exec_object(obj1);
		}
	} else error("forall","typecheck");
	unuse_object(obj1);
	unuse_object(obj2);
}

exit_op()
{
	breakflag++;
}

put()
{
	struct psobject *val = pop();
	struct psobject *key = pop();
	struct psobject *obj = pop();
	int	n;

	switch (obj->type) {
	case DICT:
		dict_define(obj,key,val);
		break;
	case ARRAY:
		if (key->type != FLOAT)
			error("put:","type check");
		n = key->val.floatval;
		if ((n < 0) || (n > obj->val.arr->len))
			error("put:","range check");
		if (obj->val.arr->val[n] != NULL) 
			unuse_object(obj->val.arr->val[n]);
		use_object(val);
		obj->val.arr->val[n] = val;
		break;
	case STRING:
		if ((key->type != FLOAT) || (val->type != FLOAT))
			error("put:","type check");
		n = key->val.floatval;
		if ((n < 0) || (n > strlen(obj->val.str)))
			error("put:","range check");
		obj->val.str[n] = (char)val->val.floatval;
		break;
	default:
		error("put:","type check");
	}
	unuse_object(obj);
	unuse_object(key);
	unuse_object(val);
}

where()
{
	struct psobject *key = pop();
	struct psobject *dict = dict_where(key);
	unuse_object(key);
	if (dict) {
		use_object(dict);
		push(dict);
		use_object(true_object);
		push(true_object);
	} else {
		key = new_object(BOOL);
		key->val.boolval = FALSE;
		use_object(key);
		push(key);
	}
}

get()
{
	struct psobject *key = pop();
	struct psobject *obj = pop();
	struct psobject *res;
	struct keyvalpair *pair;
	int n;

	if (obj->type == DICT) {
		if ((pair = dict_lookup(obj,key)) == NULL)
			error("get","undefined");
		use_object(pair->val);
		push(pair->val);
	} else if ((obj->type == ARRAY) && (key->type == FLOAT)) {
		n = key->val.floatval;
		if ((n < 0) || (n >= obj->val.arr->len))
			error("get","range check");
		use_object(obj->val.arr->val[n]);
		push(obj->val.arr->val[n]);
	} else if ((obj->type == STRING) && (key->type == FLOAT)) {
		n = key->val.floatval;
		if ((n < 0) || (n >= strlen(obj->val.str)))
			error("get","range check");
		res = new_object(FLOAT);
		res->val.floatval = obj->val.str[n];
		use_object(res);
		push(res);
	} else error("get","type check");
	unuse_object(key);
	unuse_object(obj);
}

dict()
{
	struct psobject *obj = pop_object("dict",FLOAT);
	struct psobject *dict = new_dict((int)obj->val.floatval);

	use_object(dict);
	push(dict);
	unuse_object(obj);
}

begin()
{
	struct psobject *dict = pop_object("begin",DICT);

	dict_begin(dict);
	unuse_object(dict);
}

push_currentdict()
{
	use_object(currentdict);
	push(currentdict);
}

push_currentfile()
{
	use_object(currentfile);
	push(currentfile);
}

file()
{
	struct psobject *mode = pop_object("file",STRING);
	struct psobject *path = pop_object("file",STRING), *f;
	FILE *fp = fopen(path->val.str,mode->val.str);

	if (fp == NULL)
		error("i/o error:",path->val.str);

	f = new_object(STREAM);
	f->val.file = (struct file_def *)malloc(sizeof(struct file_def));
	if (f->val.file == NULL)
		error("file:","out of memory");
	f->val.file->fp = fp;
	f->val.file->ln = 0;
	f->val.file->col = 0;
	f->val.file->name = stralloc(path->val.str);

	use_object(f);
	push(f);
	unuse_object(mode);
	unuse_object(path);
}

flip_trace()
{
	if (trace)
		printf("tracing switched OFF\n");
	else
		printf("tracing switched ON\n");
	trace = !trace;
}

print()
{
	struct psobject *obj = pop();

	print_object(stdout,obj);
	printf("\n");

	unuse_object(obj);
}

init_ops()
{
	define_operator("true",true);
	define_operator("eq",eq);
	define_operator("lt",lt);
	define_operator("nand",nand);
	define_operator("sub",sub);
	define_operator("mul",mul);
	define_operator("div",div);
	define_operator("bitshift",bitshift);
	define_operator("ln",ln);
	define_operator("sqrt",sqrt_op);
	define_operator("tan",tan_op);
	define_operator("sin",sin_op);
	define_operator("cos",cos_op);
	define_operator("floor",floor_op);
	define_operator("=",print);
	define_operator("mark",mark);
	define_operator("counttomark",counttomark);
	define_operator("array",array);
	define_operator("string",string);
	define_operator("type",type_op);
	define_operator("length",length);
	define_operator("astore",astore);
	define_operator("aload",aload);
	define_operator("cvx",cvx);
	define_operator("exec",exec);
	define_operator("put",put);
	define_operator("dict",dict);
	define_operator("begin",begin);
	define_operator("end",dict_end);
	define_operator("currentdict",push_currentdict);
	define_operator("index",index);
	define_operator("pop",pop_op);
	define_operator("clear",clear);
	define_operator("exch",exch);
	define_operator("currentfile",push_currentfile);
	define_operator("file",file);
	define_operator("trace",flip_trace);
	define_operator("ifelse",ifelse);
	define_operator("for",for_op);
	define_operator("forall",forall_op);
	define_operator("exit",exit_op);
	define_operator("where",where);
	define_operator("get",get);

	true_object = new_object(BOOL);
	true_object->val.boolval = TRUE;
	use_object(true_object);
	null_object = new_object(NILL);
	use_object(null_object);
	mark_object = new_object(MARK);
	use_object(mark_object);
}
